import doctest
import io
import unittest

import pyrtl


class TestDocTests(unittest.TestCase):
    """Test documentation examples."""

    def test_doctests(self):
        failures, tests = doctest.testmod(m=pyrtl.core)
        self.assertGreater(tests, 0)
        self.assertEqual(failures, 0)


class TestBlock(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def tearDown(self):
        pyrtl.reset_working_block()

    def test_add_wirevector_simple(self):
        w = pyrtl.WireVector(name="testwire", bitwidth=3)
        pyrtl.working_block().add_wirevector(w)
        self.assertTrue(w in pyrtl.working_block().wirevector_set)
        self.assertTrue("testwire" in pyrtl.working_block().wirevector_by_name)

    def invalid_wire(self, *args):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.working_block().add_wirevector(*args)

    def test_add_wire(self):
        self.invalid_wire(None)
        self.invalid_wire("Hi John")
        self.invalid_wire(42)

    def test_undriven_net(self):
        _ = pyrtl.WireVector(name="testwire", bitwidth=3)
        self.assertRaises(pyrtl.PyrtlError, pyrtl.working_block().sanity_check)
        pyrtl.reset_working_block()
        _ = pyrtl.Register(3)
        self.assertRaises(pyrtl.PyrtlError, pyrtl.working_block().sanity_check)
        pyrtl.reset_working_block()
        _ = pyrtl.Output(3)
        self.assertRaises(pyrtl.PyrtlError, pyrtl.working_block().sanity_check)

    def test_no_logic_net_comparisons(self):
        a = pyrtl.WireVector(bitwidth=3)
        b = pyrtl.WireVector(bitwidth=3)
        select = pyrtl.WireVector(bitwidth=3)
        outwire = pyrtl.WireVector(bitwidth=3)
        net1 = pyrtl.LogicNet(
            op="x", op_param=None, args=(select, a, b), dests=(outwire,)
        )
        net2 = pyrtl.LogicNet(
            op="x", op_param=None, args=(select, b, a), dests=(outwire,)
        )
        with self.assertRaises(pyrtl.PyrtlError):
            _ = net1 < net2
        with self.assertRaises(pyrtl.PyrtlError):
            _ = net1 <= net2
        with self.assertRaises(pyrtl.PyrtlError):
            _ = net1 > net2
        with self.assertRaises(pyrtl.PyrtlError):
            _ = net1 >= net2

    def test_logicsubset_no_op(self):
        w = pyrtl.WireVector(name="testwire1", bitwidth=1)
        v = pyrtl.WireVector(name="testwire2", bitwidth=1)
        _ = w & v
        block = pyrtl.working_block()
        self.assertEqual(block.logic_subset(None), block.logic)

    def test_sanity_check(self):
        pass

    def test_block_iterators(self):
        # testing to see that it properly runs a trivial case
        inwire = pyrtl.Input(bitwidth=1, name="inwire1")
        inwire2 = pyrtl.Input(bitwidth=1)
        inwire3 = pyrtl.Input(bitwidth=1)
        tempwire = pyrtl.WireVector()
        tempwire2 = pyrtl.WireVector()
        outwire = pyrtl.Output()

        tempwire <<= inwire | inwire2
        tempwire2 <<= ~tempwire
        outwire <<= tempwire2 & inwire3

        block = pyrtl.working_block()

        output = io.StringIO()
        for i, net in enumerate(block):
            self.assertFalse(i > 100, "Too many iterations happened")
            print(str(net), file=output)

        for net in block.logic:
            print(net, file=output)

    def test_no_memblocks(self):
        block = pyrtl.working_block()
        self.assertFalse(block.memblock_by_name)

    def test_bad_memblock_name_strict(self):
        block = pyrtl.working_block()
        with self.assertRaises(pyrtl.PyrtlError):
            _ = block.get_memblock_by_name("bad_mem", strict=True)

    def test_bad_memblock_name_none(self):
        block = pyrtl.working_block()
        mem = block.get_memblock_by_name("bad_mem")
        self.assertIsNone(mem)

    def test_same_memblock_referenced_across_multiple_operators(self):
        mem_name = "mem"
        mem = pyrtl.MemBlock(32, 5, mem_name)
        _ = mem[0]
        mem[1] <<= 42
        mem = pyrtl.working_block().get_memblock_by_name(mem_name)
        for net in pyrtl.working_block().logic:
            if net.op == "m":
                self.assertIs(net.op_param[1], mem)
            if net.op == "@":
                self.assertIs(net.op_param[1], mem)


class TestSanityCheckNet(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def tearDown(self):
        pyrtl.reset_working_block()

    def invalid_net(self, exp_message, *args):
        with self.assertRaisesRegex(pyrtl.PyrtlInternalError, exp_message):
            pyrtl.working_block().add_net(*args)

    @staticmethod
    def new_net(op="&", op_param=None, args=None, dests=None):
        if args is None or isinstance(args, int):
            args = tuple(
                pyrtl.Input(2) for i in range(args if isinstance(args, int) else 2)
            )
        if dests is None or isinstance(dests, int):

            def dest():
                return pyrtl.Register(2) if op == "r" else pyrtl.Output(2)

            dests = tuple(dest() for i in range(dests if isinstance(dests, int) else 1))
        return pyrtl.LogicNet(op=op, op_param=op_param, args=args, dests=dests)

    def test_net_make_with_not_net(self):
        self.invalid_net("net must be of type LogicNet", None)
        self.invalid_net("net must be of type LogicNet", 1)
        self.invalid_net("net must be of type LogicNet", "hi")
        self.invalid_net("net must be of type LogicNet", pyrtl.Const(2))

    def test_net_no_tuples(self):
        net = self.new_net(args=(pyrtl.Const(2)))
        self.invalid_net("LogicNet args must be tuple", net)
        net = self.new_net(dests=(pyrtl.Const(2)))
        self.invalid_net("LogicNet dests must be tuple", net)

    def test_net_odd_wires(self):
        wire = pyrtl.WireVector(2, "wire")
        net = self.new_net(args=(wire, wire))
        other_block = pyrtl.Block()
        wire._block = other_block
        self.invalid_net("net references different block", net)

        pyrtl.reset_working_block()
        wire = pyrtl.WireVector(2, "wire")
        net = self.new_net(args=(wire,))
        pyrtl.working_block().remove_wirevector(wire)
        self.invalid_net("net with unknown source", net)

    def test_net_wrong_types(self):
        inp = pyrtl.Input(2)
        const = pyrtl.Const(2)
        outp = pyrtl.Output(2)
        net = self.new_net(dests=(inp,))
        self.invalid_net("Inputs, Consts cannot be destinations", net)
        net = self.new_net(dests=(const,))
        self.invalid_net("Inputs, Consts cannot be destinations", net)
        net = self.new_net(args=(outp, outp))
        self.invalid_net("Outputs cannot be arguments", net)

        wrong_ops = ("%", "!", "a", "f", "<<", ">>", "&&", "||", "==")
        for op in wrong_ops:
            net = self.new_net(op=op)
            self.invalid_net("not from acceptable set", net)

    def test_net_wrong_num_args(self):
        for op in "w~rsm":
            net = self.new_net(op=op)
            self.invalid_net("op only allowed 1 argument", net)
        for op in "&|^n+-*<>=":
            net = self.new_net(op=op, args=3)
            self.invalid_net("op only allowed 2 arguments", net)
        for op in "x@":
            net = self.new_net(op=op, args=4)
            self.invalid_net("op only allowed 3 arguments", net)

    def test_net_wrong_bitwidth(self):
        net = self.new_net(op="x", args=tuple(pyrtl.Input(i) for i in range(1, 4)))
        self.invalid_net("args have mismatched bitwidths", net)
        net = self.new_net(op="x", args=tuple(pyrtl.Input(2) for i in range(3)))
        self.invalid_net("mux select must be a single bit", net)

        for op in "&|^n+-*<>=":
            net = self.new_net(op=op, args=(tuple(pyrtl.Input(i) for i in range(2, 4))))
            self.invalid_net("args have mismatched bitwidths", net)

        net = self.new_net(op="m", op_param=(1234, pyrtl.MemBlock(2, 3)), args=1)
        self.invalid_net("mem addrwidth mismatch", net)

        net = self.new_net(
            op="@",
            op_param=(1234, pyrtl.MemBlock(2, 2)),
            args=tuple(pyrtl.Input(i) for i in (4, 2, 1)),
        )
        self.invalid_net("mem addrwidth mismatch", net)
        net = self.new_net(
            op="@",
            op_param=(1234, pyrtl.MemBlock(2, 2)),
            args=tuple(pyrtl.Input(i) for i in (2, 4, 1)),
        )
        self.invalid_net("mem bitwidth mismatch", net)
        net = self.new_net(op="@", op_param=(1234, pyrtl.MemBlock(2, 2)), args=3)
        self.invalid_net("mem write enable must be 1 bit", net)

    def test_net_wrong_num_op_params(self):
        for op in "w~r":
            net = self.new_net(op=op, op_param=("hi", "how", "r", "u"), args=1)
            self.invalid_net("op_param should be None", net)
        for op in "&|^n+-*<>=c":
            net = self.new_net(op=op, op_param=("hi", "how", "r", "u"), args=2)
            self.invalid_net("op_param should be None", net)
        net = self.new_net(
            op="x",
            op_param=("hi", "how", "r", "u"),
            args=tuple(pyrtl.Input(i) for i in (1, 2, 2)),
        )
        self.invalid_net("op_param should be None", net)

    def test_net_wrong_op_param_select(self):
        net = self.new_net(op="s", op_param="hi", args=1)
        self.invalid_net("select op requires tuple op_param", net)
        net = self.new_net(op="s", op_param=(-2,), args=1)
        self.invalid_net("op_param out of bounds", net)
        net = self.new_net(op="s", op_param=(10,), args=1)
        self.invalid_net("op_param out of bounds", net)
        net = self.new_net(op="s", op_param=(0, 1, 3), args=1)
        self.invalid_net("op_param out of bounds", net)
        net = self.new_net(op="s", op_param=(0, True, False, {}, "hi"), args=1)
        self.invalid_net("select op_param requires ints", net)

    def test_net_wrong_op_param_mem(self):
        for op in "m@":
            net = self.new_net(
                op=op,
                op_param=[1234, pyrtl.MemBlock(1, 1)],
                args=tuple(pyrtl.Input(1) for i in range(1 if op == "m" else 3)),
            )
            self.invalid_net("mem op requires tuple op_param", net)
        for op in "m@":
            net = self.new_net(
                op=op,
                op_param=(1234, pyrtl.MemBlock(1, 1), "hi"),
                args=tuple(pyrtl.Input(1) for i in range(1 if op == "m" else 3)),
            )
            self.invalid_net("mem op requires 2 op_params in tuple", net)
        for op in "m@":
            net = self.new_net(
                op=op,
                op_param=("hi", pyrtl.MemBlock(1, 1)),
                args=tuple(pyrtl.Input(1) for i in range(1 if op == "m" else 3)),
            )
            self.invalid_net("mem op requires first operand as int", net)

        class NotMem:  # so that some earlier mem tests can work properly
            def __init__(self, bw=1, aw=1):
                self.bitwidth, self.addrwidth = bw, aw

        for op in "m@":
            net = self.new_net(
                op=op,
                op_param=(1234, NotMem()),
                args=tuple(pyrtl.Input(1) for i in range(1 if op == "m" else 3)),
            )
            self.invalid_net("mem op requires second operand of a memory type", net)

    def test_net_dest_wrong_arity_or_type(self):
        for op in "w~&|^n+-*<>=cr":
            net = self.new_net(op=op, args=1 if op in "w~r" else 2, dests=2)
            self.invalid_net("error, op only allowed 1 destination", net)
        net = self.new_net(op="s", op_param=(1,), args=1, dests=2)
        self.invalid_net("error, op only allowed 1 destination", net)
        net = self.new_net(
            op="x", args=(pyrtl.Input(1), pyrtl.Input(2), pyrtl.Input(2)), dests=2
        )
        self.invalid_net("error, op only allowed 1 destination", net)
        net = self.new_net(
            op="m", op_param=(1234, pyrtl.MemBlock(1, 2)), args=1, dests=2
        )
        self.invalid_net("error, op only allowed 1 destination", net)

        net = self.new_net(
            op="@",
            op_param=(1234, pyrtl.MemBlock(2, 2)),
            args=tuple(pyrtl.Input(i) for i in (2, 2, 1)),
        )
        self.invalid_net("mem write dest should be empty tuple", net)

        net = self.new_net(op="r", args=1, dests=(pyrtl.WireVector(2),))
        self.invalid_net("error, dest of next op should be a Register", net)

    def test_net_dest_wrong_bitwidth(self):
        for op in "w~&|^n":
            net = self.new_net(
                op=op, args=1 if op in "w~" else 2, dests=(pyrtl.Output(3),)
            )
            self.invalid_net("upper bits of destination unassigned", net)
        net = self.new_net(op="r", args=1, dests=(pyrtl.Register(3),))
        self.invalid_net("upper bits of destination unassigned", net)
        for op in "<>=":
            net = self.new_net(op=op, dests=(pyrtl.Output(2),))
            self.invalid_net("destination should be of bitwidth=1", net)
        for op in "+-":
            net = self.new_net(op=op, dests=(pyrtl.Output(4),))
            self.invalid_net("upper bits of destination unassigned", net)
        net = self.new_net(op="*", dests=(pyrtl.Output(5),))
        self.invalid_net("upper bits of destination unassigned", net)
        net = self.new_net(op="x", args=tuple(pyrtl.Input(1) for i in range(3)))
        self.invalid_net("upper bits of mux output undefined", net)
        net = self.new_net(op="c", args=3, dests=(pyrtl.Output(7),))
        self.invalid_net("upper bits of concat output undefined", net)
        net = self.new_net(op="s", args=1, op_param=(1,))
        self.invalid_net("upper bits of select output undefined", net)
        net = self.new_net(op="m", op_param=(1234, pyrtl.MemBlock(3, 2)), args=1)
        self.invalid_net("mem read dest bitwidth mismatch", net)


class TestSetWorkingBlock(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.block_a = pyrtl.Block()
        self.block_b = pyrtl.Block()

    def test_normal(self):
        pyrtl.set_working_block(self.block_a)
        self.assertIs(pyrtl.working_block(), self.block_a)
        pyrtl.set_working_block(self.block_b)
        self.assertIs(pyrtl.working_block(), self.block_b)

    def test_with_block(self):
        pyrtl.set_working_block(self.block_a)
        self.assertIs(pyrtl.working_block(), self.block_a)
        with pyrtl.set_working_block(self.block_b):
            self.assertIs(pyrtl.working_block(), self.block_b)
        self.assertIs(pyrtl.working_block(), self.block_a)

    def test_with_block_nested(self):
        self.block_c = pyrtl.Block()
        pyrtl.set_working_block(self.block_a)
        self.assertIs(pyrtl.working_block(), self.block_a)
        with pyrtl.set_working_block(self.block_b):
            self.assertIs(pyrtl.working_block(), self.block_b)
            with pyrtl.set_working_block(self.block_c):
                self.assertIs(pyrtl.working_block(), self.block_c)
            self.assertIs(pyrtl.working_block(), self.block_b)
        self.assertIs(pyrtl.working_block(), self.block_a)

    def test_with_block_exception(self):
        pyrtl.set_working_block(self.block_a)
        with self.assertRaises(pyrtl.PyrtlInternalError):
            with pyrtl.set_working_block(self.block_b):
                self.assertIs(pyrtl.working_block(), self.block_b)
                raise pyrtl.PyrtlInternalError()
        self.assertIs(pyrtl.working_block(), self.block_a)

    def test_invalid_set_wb(self):
        x = pyrtl.WireVector()
        y = 1
        pyrtl.set_working_block(self.block_a)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.set_working_block(x)
        self.assertEqual(pyrtl.working_block(), self.block_a)
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.set_working_block(y)
        self.assertEqual(pyrtl.working_block(), self.block_a)

    def test_invalid_with_set_wb(self):
        x = pyrtl.Input()
        y = True
        pyrtl.set_working_block(self.block_a)
        with self.assertRaises(pyrtl.PyrtlError):
            with pyrtl.set_working_block(x):
                pass
        self.assertEqual(pyrtl.working_block(), self.block_a)
        with self.assertRaises(pyrtl.PyrtlError):
            with pyrtl.set_working_block(y):
                pass
        self.assertEqual(pyrtl.working_block(), self.block_a)


class TestAsGraph(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def check_graph_correctness(self, w_src_graph, w_dst_graph, has_virtual=False):
        for wire, net in w_src_graph.items():
            if isinstance(wire, (pyrtl.Input, pyrtl.Const)):
                if has_virtual:
                    self.assertIs(wire, net)
                else:
                    self.fail(f"Input or Const, {wire} should not have a src")
            else:
                self.assertTrue(any(wire is w for w in net.dests))

        for wire, nets in w_dst_graph.items():
            if isinstance(wire, pyrtl.Output):
                if has_virtual:
                    self.assertEqual(len(nets), 1)
                    self.assertIs(wire, nets[0])
                else:
                    self.fail(f"Output, {wire} should not have a dst")
            else:
                for net in nets:
                    self.assertTrue(any(wire is w for w in net.args))
            self.assertEqual(len(nets), len(set(nets)))

        for net in pyrtl.working_block().logic:
            for wire in net.args:
                self.assertIn(net, w_dst_graph[wire])
            for wire in net.dests:
                self.assertIs(w_src_graph[wire], net)

    def test_as_graph_trivial(self):
        i = pyrtl.Input(1)
        o = pyrtl.Output(1)
        b = pyrtl.working_block()
        net = pyrtl.LogicNet("~", None, (i,), (o,))
        b.add_net(net)
        src_g, dst_g = b.net_connections(False)
        self.check_graph_correctness(src_g, dst_g)
        self.assertEqual(src_g[o], net)
        self.assertEqual(dst_g[i][0], net)
        self.assertEqual(len(dst_g[i]), 1)

        self.assertNotIn(i, src_g)
        self.assertNotIn(o, dst_g)

        src_g, dst_g = b.net_connections(True)
        self.check_graph_correctness(src_g, dst_g, True)
        self.assertEqual(src_g[o], net)
        self.assertEqual(dst_g[i][0], net)
        self.assertEqual(len(dst_g[i]), 1)

        self.assertIs(src_g[i], i)
        self.assertIs(dst_g[o][0], o)
        self.assertEqual(len(dst_g[o]), 1)

    def test_as_graph_2(self):
        a = pyrtl.Input(2)
        b = pyrtl.Input(2)
        c = pyrtl.Input(2)
        e = pyrtl.Output()
        f = pyrtl.Output()
        g = pyrtl.Output()

        d = a & c
        f <<= b & c
        e <<= d
        g <<= ~(d | b)

        b = pyrtl.working_block()
        src_g, dst_g = b.net_connections(False)
        self.check_graph_correctness(src_g, dst_g)

        src_g, dst_g = b.net_connections(True)
        self.check_graph_correctness(src_g, dst_g, True)

    def test_as_graph_memory(self):
        m = pyrtl.MemBlock(addrwidth=2, bitwidth=2, name="m", max_read_ports=None)
        i = pyrtl.Register(bitwidth=2, name="i")
        o = pyrtl.WireVector(bitwidth=2, name="o")
        i.next <<= i + 1
        m[i] <<= pyrtl.mux((m[i] != 0), 0, m[i])
        o <<= m[i]

        b = pyrtl.working_block()
        src_g, dst_g = b.net_connections(False)
        self.check_graph_correctness(src_g, dst_g)

        src_g, dst_g = b.net_connections(True)
        self.check_graph_correctness(src_g, dst_g, True)

    def test_as_graph_duplicate_args(self):
        a = pyrtl.Input(3)
        x = pyrtl.Input(1)
        d = pyrtl.Output()
        b = a & a
        _ = pyrtl.concat(a, a)
        m = pyrtl.MemBlock(addrwidth=3, bitwidth=3, name="m")
        m2 = pyrtl.MemBlock(addrwidth=1, bitwidth=1, name="m")
        d <<= m[a]
        m[a] <<= a
        m2[x] <<= pyrtl.MemBlock.EnabledWrite(x, x)

        b = pyrtl.working_block()
        src_g, dst_g = b.net_connections(False)
        self.check_graph_correctness(src_g, dst_g)

        src_g, dst_g = b.net_connections(True)
        self.check_graph_correctness(src_g, dst_g, True)


class TestSanityCheck(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def sanity_error(self, msg, error_type=pyrtl.PyrtlError):
        with self.assertRaisesRegex(error_type, msg):
            pyrtl.working_block().sanity_check()

    def test_missing_bitwidth(self):
        _ = pyrtl.Input()
        self.sanity_error("missing bitwidth")

    def test_duplicate_names(self):
        inp = pyrtl.Input(8, "hi")
        out = pyrtl.Output(8, "hi")
        out <<= inp
        self.sanity_error("Duplicate wire names")

    def test_unknown_wires(self):
        inp = pyrtl.Input(8, "inp")
        out = pyrtl.Output(8, "out")
        out <<= inp
        pyrtl.working_block().wirevector_set.discard(inp)
        with self.assertRaises(pyrtl.PyrtlInternalError):  # sanity_check_net()
            self.sanity_error("Unknown wires")

    def test_not_connected(self):
        _ = pyrtl.Output(8, "out")
        self.sanity_error("declared but not connected")

    def test_not_driven(self):
        w = pyrtl.WireVector(8, "w")
        out = pyrtl.Output(8, "out")
        out <<= w
        self.sanity_error("used but never driven")

    def test_inconsistent_wirevector_by_name(self):
        c = pyrtl.Const(42)
        inp = pyrtl.Input(8, "inp")
        out = pyrtl.Output(8, "out")
        out <<= inp & c
        pyrtl.working_block().wirevector_by_name["inp"] = c
        self.sanity_error(
            "inconsistent entry in wirevector_by_name", pyrtl.PyrtlInternalError
        )

    def test_missing_wire_in_wirevector_by_name(self):
        inp = pyrtl.Input(8, "inp")
        out = pyrtl.Output(8, "out")
        out <<= inp
        del pyrtl.working_block().wirevector_by_name["inp"]
        self.sanity_error(
            "Missing entries in wirevector_by_name", pyrtl.PyrtlInternalError
        )

    def test_extra_wire_in_wirevector_by_name(self):
        inp = pyrtl.Input(8, "inp")
        pyrtl.working_block().wirevector_set.discard(inp)
        self.sanity_error(
            "Unknown wires found in wirevector_by_name", pyrtl.PyrtlInternalError
        )


class TestLogicNets(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_string_format(self):
        net = pyrtl.LogicNet("+", "xx", ("arg1", "arg2"), ("dest",))
        self.assertEqual(str(net), "dest <-- + -- arg1, arg2 (xx)")

    def test_net_with_wirevectors(self):
        pass

    def test_memory_read_print(self):
        pass

    def test_memory_write_print(self):
        pass

    def test_self_equals(self):
        a = pyrtl.WireVector()
        b = pyrtl.WireVector()
        c = pyrtl.WireVector()
        net = pyrtl.LogicNet("+", "xx", (a, b), (c,))
        self.assertEqual(net, net)

    def test_comparison(self):
        net = pyrtl.LogicNet("+", "xx", ("arg1", "arg2"), ("dest",))
        with self.assertRaises(pyrtl.PyrtlError):
            _ = net < net
        with self.assertRaises(pyrtl.PyrtlError):
            _ = net <= net
        with self.assertRaises(pyrtl.PyrtlError):
            _ = net >= net
        with self.assertRaises(pyrtl.PyrtlError):
            _ = net > net

    def test_equivelence_of_same_nets(self):
        a = pyrtl.WireVector(1)
        b = pyrtl.WireVector(1)
        c = pyrtl.WireVector(1)
        net = pyrtl.LogicNet("+", "xx", (a, b), (c,))
        net2 = pyrtl.LogicNet("+", "xx", (a, b), (c,))
        self.assertIsNot(net, net2)
        self.assertEqual(net, net2)

    def assertDifferentNets(self, net1, net2):
        self.assertIsNot(net1, net2)
        self.assertNotEqual(net1, net2)
        self.assertTrue(net1 != net2)  # to test the proper working of __ne__

    def test_equivelence_of_different_nets(self):
        a = pyrtl.WireVector()
        b = pyrtl.WireVector()
        c = pyrtl.WireVector()

        n = pyrtl.LogicNet("-", "John", (a, b), (c,))
        net = pyrtl.LogicNet("+", "John", (a, b), (c,))
        net2 = pyrtl.LogicNet("+", "xx", (a, b), (c,))
        net3 = pyrtl.LogicNet("+", "xx", (b, a), (c,))
        net4 = pyrtl.LogicNet("+", "xx", (b, a, c), (c,))
        net5 = pyrtl.LogicNet("+", "xx", (b, a, c), (c, a))
        net6 = pyrtl.LogicNet("+", "xx", (b, a, c), (a,))

        self.assertDifferentNets(n, net)
        self.assertDifferentNets(net, net2)
        self.assertDifferentNets(net2, net3)
        self.assertDifferentNets(net3, net4)
        self.assertDifferentNets(net4, net5)
        self.assertDifferentNets(net4, net6)
        self.assertDifferentNets(net5, net6)

        # some extra edge cases to check
        netx_1 = pyrtl.LogicNet("+", "John", (a, a), (c,))
        netx_2 = pyrtl.LogicNet("+", "John", (a,), (c,))
        self.assertDifferentNets(netx_1, netx_2)


class TestMemAsyncCheck(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.bitwidth = 3
        self.addrwidth = 5
        self.output1 = pyrtl.Output(self.bitwidth, "output1")
        self.mem_read_address1 = pyrtl.Input(self.addrwidth, name="mem_read_address1")
        self.mem_read_address2 = pyrtl.Input(self.addrwidth, name="mem_read_address2")
        self.mem_write_address = pyrtl.Input(self.addrwidth, name="mem_write_address")
        self.mem_write_data = pyrtl.Input(self.bitwidth, name="mem_write_data")

    def tearDown(self):
        pyrtl.reset_working_block()

    def test_async_check_should_pass(self):
        memory = pyrtl.MemBlock(
            bitwidth=self.bitwidth, addrwidth=self.addrwidth, name="memory"
        )
        self.output1 <<= memory[self.mem_read_address1]
        memory[self.mem_write_address] <<= self.mem_write_data
        pyrtl.working_block().sanity_check()

    def test_async_check_should_pass_with_select(self):
        memory = pyrtl.MemBlock(
            bitwidth=self.bitwidth, addrwidth=self.addrwidth - 1, name="memory"
        )
        self.output1 <<= memory[self.mem_read_address1[0:-1]]
        pyrtl.working_block().sanity_check()

    def test_async_check_should_pass_with_cat(self):
        memory = pyrtl.MemBlock(
            bitwidth=self.bitwidth, addrwidth=self.addrwidth, name="memory"
        )
        addr = pyrtl.concat(self.mem_read_address1[0], self.mem_read_address2[0:-1])
        self.output1 <<= memory[addr]
        memory[self.mem_write_address] <<= self.mem_write_data
        pyrtl.working_block().sanity_check()

    def test_async_check_should_notpass_with_add(self):
        memory = pyrtl.MemBlock(
            bitwidth=self.bitwidth, addrwidth=self.addrwidth, name="memory"
        )
        addr = pyrtl.WireVector(self.bitwidth)
        addr <<= self.mem_read_address1 + self.mem_read_address2
        self.output1 <<= memory[addr]
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.working_block().sanity_check()


class TestNetConnections(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_net_connections_should_be_empty_at_start(self):
        src_nets, dst_nets = pyrtl.working_block().net_connections()
        self.assertDictEqual(src_nets, {})
        self.assertDictEqual(dst_nets, {})

    def test_net_connections_normal(self):
        i, j = pyrtl.input_list("i/3 j/4")
        r = pyrtl.Register(8, "r")
        o = pyrtl.Output(name="o")
        r.next <<= i * 2 + j
        o <<= r - 1

        for include_virtual in (False, True):
            src_nets, dst_nets = pyrtl.working_block().net_connections(
                include_virtual_nodes=include_virtual
            )
            if include_virtual:
                self.assertIn(i, src_nets)
                self.assertIn(j, src_nets)
                self.assertIn(o, dst_nets)
            else:
                self.assertNotIn(i, src_nets)
                self.assertNotIn(j, src_nets)
                self.assertNotIn(o, dst_nets)

            self.assertIn(o, src_nets)
            self.assertIn(r, src_nets)
            self.assertIn(i, dst_nets)
            self.assertIn(j, dst_nets)
            self.assertIn(r, dst_nets)

    def test_net_connections_with_memblock(self):
        waddr = pyrtl.Input(32, "waddr")
        raddr = pyrtl.Input(32, "raddr")
        mem = pyrtl.MemBlock(8, 32, "mem")
        data = mem[raddr]
        mem[waddr] <<= (data + pyrtl.Const(1, 8)).truncate(8)

        src_nets, dst_nets = pyrtl.working_block().net_connections()
        self.assertNotIn(data, src_nets)
        self.assertNotIn(data, dst_nets)
        self.assertIn(data.wire, src_nets)
        self.assertIn(data.wire, dst_nets)
        self.assertEqual(src_nets[data.wire].op, "m")
        self.assertEqual(dst_nets[data.wire][0].op, "+")

        with self.assertRaises(pyrtl.PyrtlError) as ex:
            _s = src_nets[data]
        self.assertEqual(
            str(ex.exception),
            "Cannot look up a _MemIndexed object's source or destination net. Try "
            "using its '.wire' attribute as the lookup key instead.",
        )

        with self.assertRaises(pyrtl.PyrtlError) as ex:
            _s = dst_nets[data]
        self.assertEqual(
            str(ex.exception),
            "Cannot look up a _MemIndexed object's source or destination net. Try "
            "using its '.wire' attribute as the lookup key instead.",
        )

    def test_wire_not_in_net_connections(self):
        w = pyrtl.WireVector()
        i = pyrtl.Input(4, "i")
        o = pyrtl.Output(name="o")
        o <<= i * 2
        src_nets, dst_nets = pyrtl.working_block().net_connections()
        with self.assertRaises(KeyError):
            src_nets[w]

        with self.assertRaises(KeyError):
            dst_nets[w]


class TestNameSanitizer(unittest.TestCase):
    def test_name_collision(self):
        # Test a sanitized name that collides with an unsanitized name.
        sanitizer = pyrtl.core._NameSanitizer(
            identifier_regex_str=pyrtl.core._py_regex,
            internal_prefix="sanitized",
        )
        # This name should be sanitized by replacing the dot with an underscore.
        sanitized_dot = sanitizer.make_valid_string("foo.bar")
        self.assertEqual(sanitized_dot, "foo_bar")
        self.assertEqual(sanitizer["foo.bar"], "foo_bar")

        # This name collides with the sanitized name we just generated, so the first
        # internal_index should be appended.
        sanitized_underscore = sanitizer.make_valid_string("foo_bar")
        self.assertEqual(sanitized_underscore, "foo_bar0")
        self.assertEqual(sanitizer["foo_bar"], "foo_bar0")

        # This name does not require sanitization, but it will collide with the next
        # foo.bar variant.
        sanitized_one = sanitizer.make_valid_string("foo_bar1")
        self.assertEqual(sanitized_one, "foo_bar1")
        self.assertEqual(sanitizer["foo_bar1"], "foo_bar1")

        # Attempting to sanitize foo!bar by appending internal_index collides with the
        # name we just registered. _NameSanitizer should give up and use
        # internal_prefix.
        sanitized_bang = sanitizer.make_valid_string("foo!bar")
        self.assertEqual(sanitized_bang, "sanitized2")
        self.assertEqual(sanitizer["foo!bar"], "sanitized2")


if __name__ == "__main__":
    unittest.main()



import doctest
import random
import unittest

import pyrtl
from pyrtl.rtllib.barrel import Direction, barrel_shifter


class TestDocTests(unittest.TestCase):
    """Test documentation examples."""

    def test_doctests(self):
        failures, tests = doctest.testmod(m=pyrtl.rtllib.barrel)
        self.assertGreater(tests, 0)
        self.assertEqual(failures, 0)


class TestBarrel(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.inp_val = pyrtl.Input(8, "inp_val")
        self.inp_shift = pyrtl.Input(2, "inp_shift")
        self.out_zeros = pyrtl.Output(18, "out_zeros")
        self.out_ones = pyrtl.Output(18, "out_ones")

    def test_shift_left(self):
        random.seed(777906373)
        self.out_zeros <<= barrel_shifter(
            self.inp_val, bit_in=0, direction=Direction.LEFT, shift_dist=self.inp_shift
        )
        self.out_ones <<= barrel_shifter(
            self.inp_val, bit_in=1, direction=Direction.LEFT, shift_dist=self.inp_shift
        )

        sim = pyrtl.Simulation()
        vals = [random.randint(0, 20) for v in range(20)]
        shifts = [random.randint(0, 3) for s in range(20)]
        for i in range(len(vals)):
            sim.step({self.inp_val: vals[i], self.inp_shift: shifts[i]})
            base_sum = vals[i] * pow(2, shifts[i])
            self.assertEqual(sim.inspect(self.out_zeros), base_sum)
            self.assertEqual(
                sim.inspect(self.out_ones), base_sum + pow(2, shifts[i]) - 1
            )

    def test_shift_right(self):
        random.seed(777906374)
        self.out_zeros <<= barrel_shifter(
            self.inp_val, bit_in=0, direction=Direction.RIGHT, shift_dist=self.inp_shift
        )
        self.out_ones <<= barrel_shifter(
            self.inp_val, bit_in=1, direction=Direction.RIGHT, shift_dist=self.inp_shift
        )

        sim = pyrtl.Simulation()
        vals = [random.randint(0, 20) for v in range(20)]
        shifts = [random.randint(0, 3) for s in range(20)]
        for i in range(len(vals)):
            sim.step({self.inp_val: vals[i], self.inp_shift: shifts[i]})
            base_sum = int(vals[i] / pow(2, shifts[i]))
            self.assertEqual(
                sim.inspect(self.out_zeros), base_sum, f"failed on value {vals[i]}"
            )
            extra_sum = sum(
                [pow(2, len(self.inp_val) - b - 1) for b in range(shifts[i])]
            )
            self.assertEqual(
                sim.inspect(self.out_ones),
                base_sum + extra_sum,
                f"failed on value {vals[i]}",
            )


if __name__ == "__main__":
    unittest.main()



# # Introduction to Hardware Design

# This code works through the hardware design process with the audience of software
# developers more in mind. We start with the simple problem of designing a Fibonacci
# sequence calculator (http://oeis.org/A000045).
import pyrtl


# ## Software Fibonacci implementation
def software_fibonacci(n: int):
    """An ordinary Python function to return the `n`th Fibonacci number."""
    a = 0
    b = 1
    for _ in range(n):
        a, b = b, a + b
    return a


# `software_fibonacci` is an iterative Fibonacci implementation. It repeatedly adds `a`
# and `b` to calculate the `n`th number in the sequence.
print("n   software_fibonacci(n)")
print("─────────────────────────")
for n in range(10):
    print(n, " ", software_fibonacci(n))


# ## Hardware Fibonacci implementation, first attempt
#
# So let's convert this into some hardware that computes the same thing. Our first go
# will be to just replace the `0` and `1` with `WireVectors` to see what happens.
def attempt1_hardware_fibonacci(n: int) -> pyrtl.WireVector:
    a = pyrtl.Const(0)
    b = pyrtl.Const(1)
    for _ in range(n):
        a, b = b, a + b
    return a


# The above looks really nice but does not really represent a hardware implementation
# of Fibonacci. Let's reason through the code, line by line, to figure out what
# it would actually build.
#
#     a = pyrtl.Const(0)      # ← This makes a `WireVector` with `bitwidth=1` that is
#                             #   driven by a zero.  Thus `a` is a `WireVector`.  Seems
#                             #   good.
#     b = pyrtl.Const(1)      # ← Just like above, `b` is a `WireVector` driven by `1`.
#        for i in range(n):   # ← Okay, here is where things start to go off the rails a
#                             #   bit. This says to perform the following code `n`
#                             #   times, but the value `n` is passed as an argument and
#                             #   is not something that is evaluated in the hardware; it
#                             #   is evaluated when you run the PyRTL program which
#                             #   generates (or more specifically elaborates) the
#                             #   hardware. Thus the hardware we are building will have
#                             #   the value of `n` built into the hardware and won't
#                             #   actually be a run-time parameter. Loops are useful for
#                             #   building large repetitive hardware structures, but
#                             #   they CAN'T be used to represent hardware that should
#                             #   do a computation iteratively. Instead we will need
#                             #   some `Registers` to build a state machine.
#            a, b = b, a + b  # ← Let's break this apart. In the first cycle, `a` is
#                             #   `Const(0) and `b` is `Const(1)`, so `(a + b)` builds
#                             #   an adder with `Const(0)` and `Const(1)` as inputs.
#                             #   So `(b, a + b)` in the first iteration evaluates to:
#                             #   `(Const(1), result_of_adding(Const(0), Const(1)))`.
#                             #   At the end of the first iteration, `a` and `b` refer
#                             #   to those two constant values. In each following
#                             #   iteration more adders are built and the Python
#                             #   variables `a` and `b` are bound to larger and larger
#                             #   trees of adders. But all the adder inputs are
#                             #   constants!
#        return a             # ← The final thing that is returned then is the last
#                             #   output from this tree of adders which all have
#                             #   `Consts` as inputs. Thus this hardware is hard-wired
#                             #   to produce only the exact the value of the `n`th
#                             #   Fibonacci number, where `n` is hard-coded into the
#                             #   design! This is most likely not what you want.
#
# ## Hardware Fibonacci implementation, second attempt
#
# So let's try a different approach. Let's specify two registers (`a` and `b`) and then
# we can update those values as we iteratively compute the `n`th Fibonacci number cycle
# by cycle.
def attempt2_hardware_fibonacci(bitwidth: int) -> pyrtl.WireVector:
    a = pyrtl.Register(bitwidth, "a")
    b = pyrtl.Register(bitwidth, "b")

    a.next <<= b
    b.next <<= a + b

    return a


# This is looking much better. Two registers, `a` and `b`, store the values from which
# we can compute the series. The line `a.next <<= b` means that the value of `a` in the
# next cycle should be simply be `b` from the current cycle. The line `b.next <<= a + b`
# says to build an adder with inputs of `a` and `b` from the current cycle and assign
# the value to `b` in the next cycle. A visual representation of the hardware built is
# as such:
#
#        ┌─────┐     ┌─────┐
#        │     │     │     │
#        ▼     │     ▼     │
#     ▕▔▔▔▔▔▏  │  ▕▔▔▔▔▔▏  │
#     ▕  a  ▏  │  ▕  b  ▏  │
#     ▕▁▁▁▁▁▏  │  ▕▁▁▁▁▁▏  │
#        │     │     │     │
#        │     └─────┤     │
#        │           │     │
#        ▼           ▼     │
#      ╲▔▔▔▔▔▔▔▔▔▔▔▔▔▔▔╱   │
#       ╲    adder    ╱    │
#        ╲▁▁▁▁▁▁▁▁▁▁▁╱     │
#              │           │
#              └───────────┘
#
# Note that in the picture the register `a` and `b` each have a `WireVector` which is
# the current value (shown flowing out of the bottom of the register) and an "input"
# which is giving the value that should be the value of the register in the following
# cycle (shown flowing into the top of the register) which are `b` and `a + b`
# respectively. When we say `return a` what we are returning is a reference to the
# register `a` in the picture above.
#
# ## Hardware Fibonacci implementation, third attempt
#
# Of course one problem is that we don't know when we are done. How do we know we
# reached the `n`th number in the sequence? Well, we need to add a `Register` to keep
# count and see if we are done.
def attempt3_hardware_fibonacci(
    n: pyrtl.WireVector,
) -> tuple[pyrtl.WireVector, pyrtl.WireVector]:
    a = pyrtl.Register(bitwidth=n.bitwidth, name="a")
    b = pyrtl.Register(bitwidth=n.bitwidth, name="b")
    counter = pyrtl.Register(bitwidth=n.bitwidth, name="counter")

    counter.next <<= counter + 1
    a.next <<= b
    b.next <<= a + b

    return a, counter == n


# This is very similar to the example before, except that now we have a register
# `counter` which keeps track of the iteration that we are on:
#     counter.next <<= counter + 1
# The function now returns two values, a reference to the register `a` and a reference
# to a single bit that tells us if we are done. That bit is calculated by comparing
# `counter` to the to a WireVector `n` that is passed in to see if they are the same.
#
# ## Hardware Fibonacci implementation, fourth attempt
#
# Finally, we need a way to indicate that we want a new Fibonacci number. We'll add
# another input, `start`, which when high sets our `local_n` register and resets the
# others. Now our ending condition occurs when the current iteration `counter` is equal
# to the locally stored `local_n`.
def attempt4_hardware_fibonacci(
    n: pyrtl.WireVector, start: pyrtl.WireVector
) -> tuple[pyrtl.WireVector, pyrtl.WireVector]:
    a = pyrtl.Register(bitwidth=n.bitwidth, name="a")
    b = pyrtl.Register(bitwidth=n.bitwidth, name="b")
    counter = pyrtl.Register(bitwidth=n.bitwidth, name="counter")
    local_n = pyrtl.Register(bitwidth=n.bitwidth, name="local_n")
    local_done = pyrtl.WireVector(bitwidth=1, name="local_done")

    with pyrtl.conditional_assignment:
        with start:
            local_n.next |= n
            counter.next |= 0
            a.next |= 0
            b.next |= 1
        with pyrtl.otherwise:
            counter.next |= counter + 1
            a.next |= b
            b.next |= a + b
    local_done <<= counter == local_n
    return a, local_done


# ## Simulating the hardware Fibonacci implementation
#
# This is now far enough along that we can simulate the design and see what happens. We
# begin by connecting our input and output wires to the implementation, stepping once
# with the `start` signal high to signify we're starting a new Fibonacci request, and
# then continuing to `step()` until `done` is asserted. Note that although the Fibonacci
# implementation only uses the value of `n` when `start` is high, we must still provide
# a value for `n` (and all other `Inputs` tracked by the `Simulation`) for each
# `step()`.
print("\nHardware Fibonacci Simulation:")
n = pyrtl.Input(bitwidth=8, name="n")
start = pyrtl.Input(bitwidth=1, name="start")
fib_out = pyrtl.Output(bitwidth=8, name="fib")
done_out = pyrtl.Output(bitwidth=1, name="done")

fib, done = attempt4_hardware_fibonacci(n, start)
fib_out <<= fib
done_out <<= done

sim = pyrtl.Simulation()

sim.step({"n": 7, "start": 1})

sim.step({"n": 0, "start": 0})
while not sim.inspect("done"):
    sim.step({"n": 0, "start": 0})

sim.tracer.render_trace(
    trace_list=["n", "start", "counter", "fib", "done"], repr_func=int
)

assert sim.inspect("fib") == software_fibonacci(7)



import numbers

import pyrtl
from pyrtl.rtllib import libutils

# TODO:
# 2) All ROMs should be synchronous.  This should be easy once (3) is completed
# 3) Right now decryption generates one GIANT combinatorial block. Instead
#    it should generate one of 2 options -- Either an iterative design or a
#    pipelined design.  Both will add registers between each round of AES
# 5) a single "aes-unit" combining encryption and decryption (without making
#    full independent hardware units) would be a plus as well


class AES:
    """A class for building a PyRTL AES circuit.

    Currently this class only supports 128 bit AES encryption/decryption.

    Example::

        import pyrtl
        from pyrtl.rtllib.aes import AES

        aes = AES()
        plaintext = pyrtl.Input(bitwidth=128, name='aes_plaintext')
        key = pyrtl.Input(bitwidth=128, name='aes_key')
        aes_ciphertext = pyrtl.Output(bitwidth=128, name='aes_ciphertext')
        reset = pyrtl.Input(1, name='reset')
        ready = pyrtl.Output(1, name='ready')

        ready_out, aes_cipher = aes.encrypt_state_m(plaintext, key, reset)
        ready <<= ready_out
        aes_ciphertext <<= aes_cipher

        sim = pyrtl.Simulation()
        sim.step ({
            'aes_plaintext': 0x00112233445566778899aabbccddeeff,
            'aes_key': 0x000102030405060708090a0b0c0d0e0f,
            'reset': 1
        })
        for cycle in range(1,10):
            sim.step ({
                'aes_plaintext': 0x00112233445566778899aabbccddeeff,
                'aes_key': 0x000102030405060708090a0b0c0d0e0f,
                'reset': 0
            })
        sim.tracer.render_trace(symbol_len=40, segment_size=1)
    """

    def __init__(self):
        self.memories_built = False
        self._key_len = 128

    def encryption(
        self, plaintext: pyrtl.WireVector, key: pyrtl.WireVector
    ) -> pyrtl.WireVector:
        """Builds a single cycle AES Encryption circuit.

        :param plaintext: Text to encrypt.
        :param key: AES key to use to encrypt.

        :return: A :class:`.WireVector` containing the ciphertext.
        """
        if len(plaintext) != self._key_len:
            msg = "Ciphertext length is invalid"
            raise pyrtl.PyrtlError(msg)
        if len(key) != self._key_len:
            msg = "key length is invalid"
            raise pyrtl.PyrtlError(msg)

        key_list = self._key_gen(key)
        t = self._add_round_key(plaintext, key_list[0])

        for round in range(1, 11):
            t = self._sub_bytes(t)
            t = self._shift_rows(t)
            if round != 10:
                t = self._mix_columns(t)
            t = self._add_round_key(t, key_list[round])
        return t

    def encrypt_state_m(
        self,
        plaintext_in: pyrtl.WireVector,
        key_in: pyrtl.WireVector,
        reset: pyrtl.WireVector,
    ) -> tuple[pyrtl.WireVector, pyrtl.WireVector]:
        """Builds a multiple cycle AES Encryption state machine circuit.

        :param plaintext: Text to encrypt.
        :param key: AES key to use to encrypt.
        :param reset: a one bit signal telling the state machine to reset and accept the
            current ``plaintext`` and ``key``.

        :return: ``(ready, cipher_text)``: ``ready`` is a one bit signal showing that
                 the encryption result (``cipher_text``) has been calculated.
        """
        if len(key_in) != len(plaintext_in):
            msg = "AES key and plaintext should be the same length"
            raise pyrtl.PyrtlError(msg)

        plain_text, key = (pyrtl.Register(len(plaintext_in)) for i in range(2))
        key_exp_in, add_round_in = (
            pyrtl.WireVector(len(plaintext_in)) for i in range(2)
        )

        counter = pyrtl.Register(4, "counter")
        round = pyrtl.WireVector(4, "round")
        counter.next <<= round
        sub_out = self._sub_bytes(plain_text)
        shift_out = self._shift_rows(sub_out)
        mix_out = self._mix_columns(shift_out)
        key_out = self._key_expansion(key, counter)
        add_round_out = self._add_round_key(add_round_in, key_exp_in)
        with pyrtl.conditional_assignment:
            with reset == 1:
                round |= 0
                key_exp_in |= key_in  # to lower the number of cycles
                plain_text.next |= add_round_out
                key.next |= key_in
                add_round_in |= plaintext_in

            with counter == 10:  # keep everything the same
                round |= counter
                plain_text.next |= plain_text

            with pyrtl.otherwise:  # running through AES
                round |= counter + 1
                key_exp_in |= key_out
                plain_text.next |= add_round_out
                key.next |= key_out
                with counter == 9:
                    add_round_in |= shift_out
                with pyrtl.otherwise:
                    add_round_in |= mix_out

        ready = counter == 10
        return ready, plain_text

    def decryption(
        self, ciphertext: pyrtl.WireVector, key: pyrtl.WireVector
    ) -> pyrtl.WireVector:
        """Builds a single cycle AES Decryption circuit.

        :param ciphertext: Data to decrypt.
        :param key: AES key to use to encrypt (AES is symmetric).

        :return: A :class:`.WireVector` containing the plaintext.
        """
        if len(ciphertext) != self._key_len:
            msg = "Ciphertext length is invalid"
            raise pyrtl.PyrtlError(msg)
        if len(key) != self._key_len:
            msg = "key length is invalid"
            raise pyrtl.PyrtlError(msg)
        key_list = self._key_gen(key)
        t = self._add_round_key(ciphertext, key_list[10])

        for round in range(1, 11):
            t = self._inv_shift_rows(t)
            t = self._sub_bytes(t, True)
            t = self._add_round_key(t, key_list[10 - round])
            if round != 10:
                t = self._mix_columns(t, True)

        return t

    def decryption_statem(
        self,
        ciphertext_in: pyrtl.WireVector,
        key_in: pyrtl.WireVector,
        reset: pyrtl.WireVector,
    ) -> tuple[pyrtl.WireVector, pyrtl.WireVector]:
        """Builds a multiple cycle AES Decryption state machine circuit.

        :param ciphertext: Data to decrypt.
        :param key: AES key to use to encrypt (AES is symmetric).
        :param reset: a one bit signal telling the state machine to reset and accept the
            current plaintext and key.

        :return: ``(ready, plain_text)``: ``ready`` is a one bit signal showing that the
                 decryption result (``plain_text``) has been calculated.
        """
        if len(key_in) != len(ciphertext_in):
            msg = "AES key and ciphertext should be the same length"
            raise pyrtl.PyrtlError(msg)

        cipher_text, key = (pyrtl.Register(len(ciphertext_in)) for i in range(2))
        key_exp_in, add_round_in = (
            pyrtl.WireVector(len(ciphertext_in)) for i in range(2)
        )

        # this is not part of the state machine as we need the keys in reverse order...
        reversed_key_list = reversed(self._key_gen(key_exp_in))

        counter = pyrtl.Register(4, "counter")
        round = pyrtl.WireVector(4)
        counter.next <<= round

        inv_shift = self._inv_shift_rows(cipher_text)
        inv_sub = self._sub_bytes(inv_shift, True)
        key_out = pyrtl.mux(round, *reversed_key_list, default=0)
        add_round_out = self._add_round_key(add_round_in, key_out)
        inv_mix_out = self._mix_columns(add_round_out, True)

        with pyrtl.conditional_assignment:
            with reset == 1:
                round |= 0
                key.next |= key_in
                key_exp_in |= key_in  # to lower the number of cycles needed
                cipher_text.next |= add_round_out
                add_round_in |= ciphertext_in

            with counter == 10:  # keep everything the same
                round |= counter
                cipher_text.next |= cipher_text

            with pyrtl.otherwise:  # running through AES
                round |= counter + 1

                key.next |= key
                key_exp_in |= key
                add_round_in |= inv_sub
                with counter == 9:
                    cipher_text.next |= add_round_out
                with pyrtl.otherwise:
                    cipher_text.next |= inv_mix_out

        ready = counter == 10
        return ready, cipher_text

    def _key_gen(self, key):
        keys = [key]
        for enc_round in range(10):
            key = self._key_expansion(key, enc_round)
            keys.append(key)
        return keys

    def _key_expansion(self, old_key, key_expand_round):
        self._build_memories_if_not_exists()
        w = libutils.partition_wire(old_key, 32)
        x = [w[3] ^ self._g(w[0], key_expand_round)]
        x.insert(0, x[0] ^ w[2])
        x.insert(0, x[0] ^ w[1])
        x.insert(0, x[0] ^ w[0])
        return pyrtl.concat_list(x)

    def _g(self, word, key_expand_round):
        """
        One-byte left circular rotation, substitution of each byte
        """
        self._build_memories_if_not_exists()
        a = libutils.partition_wire(word, 8)
        sub = [self.sbox[a[index]] for index in (3, 0, 1, 2)]
        if isinstance(key_expand_round, numbers.Number):
            rcon_data = self._rcon_data[key_expand_round + 1]  # int value
        else:
            rcon_data = self.rcon[key_expand_round + 1]
        sub[3] = sub[3] ^ rcon_data
        return pyrtl.concat_list(sub)

    def _sub_bytes(self, in_vector, inverse=False):
        self._build_memories_if_not_exists()
        subbed = [
            self.inv_sbox[byte] if inverse else self.sbox[byte]
            for byte in libutils.partition_wire(in_vector, 8)
        ]
        return pyrtl.concat_list(subbed)

    @staticmethod
    def _inv_shift_rows(in_vector):
        a = libutils.partition_wire(in_vector, 8)
        return pyrtl.concat_list(
            (
                a[12],
                a[9],
                a[6],
                a[3],
                a[0],
                a[13],
                a[10],
                a[7],
                a[4],
                a[1],
                a[14],
                a[11],
                a[8],
                a[5],
                a[2],
                a[15],
            )
        )

    @staticmethod
    def _shift_rows(in_vector):
        a = libutils.partition_wire(in_vector, 8)
        return pyrtl.concat_list(
            (
                a[4],
                a[9],
                a[14],
                a[3],
                a[8],
                a[13],
                a[2],
                a[7],
                a[12],
                a[1],
                a[6],
                a[11],
                a[0],
                a[5],
                a[10],
                a[15],
            )
        )

    def _galois_mult(self, c, mult_table):
        if mult_table == 1:
            return c
        return self._galois_mults[mult_table][c]

    def _mix_columns(self, in_vector, inverse=False):
        self._build_memories_if_not_exists()
        igm_mults = [14, 9, 13, 11] if inverse else [2, 1, 1, 3]
        subgroups = libutils.partition_wire(in_vector, 32)
        return pyrtl.concat_list(
            [self._mix_col_subgroup(sg, igm_mults) for sg in subgroups]
        )

    def _mix_col_subgroup(self, in_vector, gm_multipliers):
        def _mix_single(index):
            mult_items = [
                self._galois_mult(a[(index + loc) % 4], mult_table)
                for loc, mult_table in enumerate(gm_multipliers)
            ]
            return mult_items[0] ^ mult_items[1] ^ mult_items[2] ^ mult_items[3]

        a = libutils.partition_wire(in_vector, 8)
        return pyrtl.concat_list([_mix_single(index) for index in range(len(a))])

    @staticmethod
    def _add_round_key(t, key):
        return t ^ key

    def _build_memories_if_not_exists(self):
        if not self.memories_built:
            self._build_memories()

    def _build_memories(self):
        def build_mem(data):
            return pyrtl.RomBlock(
                bitwidth=8,
                addrwidth=8,
                romdata=data,
                build_new_roms=True,
                asynchronous=True,
            )

        self.sbox = build_mem(self._sbox_data)
        self.inv_sbox = build_mem(self._inv_sbox_data)
        self.rcon = build_mem(self._rcon_data)
        self.GM2 = build_mem(self._GM2_data)
        self.GM3 = build_mem(self._GM3_data)
        self.GM9 = build_mem(self._GM9_data)
        self.GM11 = build_mem(self._GM11_data)
        self.GM13 = build_mem(self._GM13_data)
        self.GM14 = build_mem(self._GM14_data)
        self._galois_mults = {
            3: self.GM3,
            2: self.GM2,
            9: self.GM9,
            11: self.GM11,
            13: self.GM13,
            14: self.GM14,
        }
        self.memories_built = True

    _sbox_data = libutils.str_to_int_array("""
        63 7c 77 7b f2 6b 6f c5 30 01 67 2b fe d7 ab 76 ca 82 c9 7d fa 59 47 f0
        ad d4 a2 af 9c a4 72 c0 b7 fd 93 26 36 3f f7 cc 34 a5 e5 f1 71 d8 31 15
        04 c7 23 c3 18 96 05 9a 07 12 80 e2 eb 27 b2 75 09 83 2c 1a 1b 6e 5a a0
        52 3b d6 b3 29 e3 2f 84 53 d1 00 ed 20 fc b1 5b 6a cb be 39 4a 4c 58 cf
        d0 ef aa fb 43 4d 33 85 45 f9 02 7f 50 3c 9f a8 51 a3 40 8f 92 9d 38 f5
        bc b6 da 21 10 ff f3 d2 cd 0c 13 ec 5f 97 44 17 c4 a7 7e 3d 64 5d 19 73
        60 81 4f dc 22 2a 90 88 46 ee b8 14 de 5e 0b db e0 32 3a 0a 49 06 24 5c
        c2 d3 ac 62 91 95 e4 79 e7 c8 37 6d 8d d5 4e a9 6c 56 f4 ea 65 7a ae 08
        ba 78 25 2e 1c a6 b4 c6 e8 dd 74 1f 4b bd 8b 8a 70 3e b5 66 48 03 f6 0e
        61 35 57 b9 86 c1 1d 9e e1 f8 98 11 69 d9 8e 94 9b 1e 87 e9 ce 55 28 df
        8c a1 89 0d bf e6 42 68 41 99 2d 0f b0 54 bb 16
        """)

    _inv_sbox_data = libutils.str_to_int_array("""
        52 09 6a d5 30 36 a5 38 bf 40 a3 9e 81 f3 d7 fb 7c e3 39 82 9b 2f ff 87
        34 8e 43 44 c4 de e9 cb 54 7b 94 32 a6 c2 23 3d ee 4c 95 0b 42 fa c3 4e
        08 2e a1 66 28 d9 24 b2 76 5b a2 49 6d 8b d1 25 72 f8 f6 64 86 68 98 16
        d4 a4 5c cc 5d 65 b6 92 6c 70 48 50 fd ed b9 da 5e 15 46 57 a7 8d 9d 84
        90 d8 ab 00 8c bc d3 0a f7 e4 58 05 b8 b3 45 06 d0 2c 1e 8f ca 3f 0f 02
        c1 af bd 03 01 13 8a 6b 3a 91 11 41 4f 67 dc ea 97 f2 cf ce f0 b4 e6 73
        96 ac 74 22 e7 ad 35 85 e2 f9 37 e8 1c 75 df 6e 47 f1 1a 71 1d 29 c5 89
        6f b7 62 0e aa 18 be 1b fc 56 3e 4b c6 d2 79 20 9a db c0 fe 78 cd 5a f4
        1f dd a8 33 88 07 c7 31 b1 12 10 59 27 80 ec 5f 60 51 7f a9 19 b5 4a 0d
        2d e5 7a 9f 93 c9 9c ef a0 e0 3b 4d ae 2a f5 b0 c8 eb bb 3c 83 53 99 61
        17 2b 04 7e ba 77 d6 26 e1 69 14 63 55 21 0c 7d
        """)

    _rcon_data = libutils.str_to_int_array("""
        8d 01 02 04 08 10 20 40 80 1b 36 6c d8 ab 4d 9a 2f 5e bc 63 c6 97 35 6a
        d4 b3 7d fa ef c5 91 39 72 e4 d3 bd 61 c2 9f 25 4a 94 33 66 cc 83 1d 3a
        74 e8 cb 8d 01 02 04 08 10 20 40 80 1b 36 6c d8 ab 4d 9a 2f 5e bc 63 c6
        97 35 6a d4 b3 7d fa ef c5 91 39 72 e4 d3 bd 61 c2 9f 25 4a 94 33 66 cc
        83 1d 3a 74 e8 cb 8d 01 02 04 08 10 20 40 80 1b 36 6c d8 ab 4d 9a 2f 5e
        bc 63 c6 97 35 6a d4 b3 7d fa ef c5 91 39 72 e4 d3 bd 61 c2 9f 25 4a 94
        33 66 cc 83 1d 3a 74 e8 cb 8d 01 02 04 08 10 20 40 80 1b 36 6c d8 ab 4d
        9a 2f 5e bc 63 c6 97 35 6a d4 b3 7d fa ef c5 91 39 72 e4 d3 bd 61 c2 9f
        25 4a 94 33 66 cc 83 1d 3a 74 e8 cb 8d 01 02 04 08 10 20 40 80 1b 36 6c
        d8 ab 4d 9a 2f 5e bc 63 c6 97 35 6a d4 b3 7d fa ef c5 91 39 72 e4 d3 bd
        61 c2 9f 25 4a 94 33 66 cc 83 1d 3a 74 e8 cb 8d
        """)

    # Galois Multiplication tables for 2, 3, 9, 11, 13, and 14.

    _GM2_data = libutils.str_to_int_array("""
        00 02 04 06 08 0a 0c 0e 10 12 14 16 18 1a 1c 1e 20 22 24 26 28 2a 2c 2e
        30 32 34 36 38 3a 3c 3e 40 42 44 46 48 4a 4c 4e 50 52 54 56 58 5a 5c 5e
        60 62 64 66 68 6a 6c 6e 70 72 74 76 78 7a 7c 7e 80 82 84 86 88 8a 8c 8e
        90 92 94 96 98 9a 9c 9e a0 a2 a4 a6 a8 aa ac ae b0 b2 b4 b6 b8 ba bc be
        c0 c2 c4 c6 c8 ca cc ce d0 d2 d4 d6 d8 da dc de e0 e2 e4 e6 e8 ea ec ee
        f0 f2 f4 f6 f8 fa fc fe 1b 19 1f 1d 13 11 17 15 0b 09 0f 0d 03 01 07 05
        3b 39 3f 3d 33 31 37 35 2b 29 2f 2d 23 21 27 25 5b 59 5f 5d 53 51 57 55
        4b 49 4f 4d 43 41 47 45 7b 79 7f 7d 73 71 77 75 6b 69 6f 6d 63 61 67 65
        9b 99 9f 9d 93 91 97 95 8b 89 8f 8d 83 81 87 85 bb b9 bf bd b3 b1 b7 b5
        ab a9 af ad a3 a1 a7 a5 db d9 df dd d3 d1 d7 d5 cb c9 cf cd c3 c1 c7 c5
        fb f9 ff fd f3 f1 f7 f5 eb e9 ef ed e3 e1 e7 e5
        """)

    _GM3_data = libutils.str_to_int_array("""
        00 03 06 05 0c 0f 0a 09 18 1b 1e 1d 14 17 12 11 30 33 36 35 3c 3f 3a 39
        28 2b 2e 2d 24 27 22 21 60 63 66 65 6c 6f 6a 69 78 7b 7e 7d 74 77 72 71
        50 53 56 55 5c 5f 5a 59 48 4b 4e 4d 44 47 42 41 c0 c3 c6 c5 cc cf ca c9
        d8 db de dd d4 d7 d2 d1 f0 f3 f6 f5 fc ff fa f9 e8 eb ee ed e4 e7 e2 e1
        a0 a3 a6 a5 ac af aa a9 b8 bb be bd b4 b7 b2 b1 90 93 96 95 9c 9f 9a 99
        88 8b 8e 8d 84 87 82 81 9b 98 9d 9e 97 94 91 92 83 80 85 86 8f 8c 89 8a
        ab a8 ad ae a7 a4 a1 a2 b3 b0 b5 b6 bf bc b9 ba fb f8 fd fe f7 f4 f1 f2
        e3 e0 e5 e6 ef ec e9 ea cb c8 cd ce c7 c4 c1 c2 d3 d0 d5 d6 df dc d9 da
        5b 58 5d 5e 57 54 51 52 43 40 45 46 4f 4c 49 4a 6b 68 6d 6e 67 64 61 62
        73 70 75 76 7f 7c 79 7a 3b 38 3d 3e 37 34 31 32 23 20 25 26 2f 2c 29 2a
        0b 08 0d 0e 07 04 01 02 13 10 15 16 1f 1c 19 1a
        """)

    _GM9_data = libutils.str_to_int_array("""
        00 09 12 1b 24 2d 36 3f 48 41 5a 53 6c 65 7e 77 90 99 82 8b b4 bd a6 af
        d8 d1 ca c3 fc f5 ee e7 3b 32 29 20 1f 16 0d 04 73 7a 61 68 57 5e 45 4c
        ab a2 b9 b0 8f 86 9d 94 e3 ea f1 f8 c7 ce d5 dc 76 7f 64 6d 52 5b 40 49
        3e 37 2c 25 1a 13 08 01 e6 ef f4 fd c2 cb d0 d9 ae a7 bc b5 8a 83 98 91
        4d 44 5f 56 69 60 7b 72 05 0c 17 1e 21 28 33 3a dd d4 cf c6 f9 f0 eb e2
        95 9c 87 8e b1 b8 a3 aa ec e5 fe f7 c8 c1 da d3 a4 ad b6 bf 80 89 92 9b
        7c 75 6e 67 58 51 4a 43 34 3d 26 2f 10 19 02 0b d7 de c5 cc f3 fa e1 e8
        9f 96 8d 84 bb b2 a9 a0 47 4e 55 5c 63 6a 71 78 0f 06 1d 14 2b 22 39 30
        9a 93 88 81 be b7 ac a5 d2 db c0 c9 f6 ff e4 ed 0a 03 18 11 2e 27 3c 35
        42 4b 50 59 66 6f 74 7d a1 a8 b3 ba 85 8c 97 9e e9 e0 fb f2 cd c4 df d6
        31 38 23 2a 15 1c 07 0e 79 70 6b 62 5d 54 4f 46
        """)

    _GM11_data = libutils.str_to_int_array("""
        00 0b 16 1d 2c 27 3a 31 58 53 4e 45 74 7f 62 69 b0 bb a6 ad 9c 97 8a 81
        e8 e3 fe f5 c4 cf d2 d9 7b 70 6d 66 57 5c 41 4a 23 28 35 3e 0f 04 19 12
        cb c0 dd d6 e7 ec f1 fa 93 98 85 8e bf b4 a9 a2 f6 fd e0 eb da d1 cc c7
        ae a5 b8 b3 82 89 94 9f 46 4d 50 5b 6a 61 7c 77 1e 15 08 03 32 39 24 2f
        8d 86 9b 90 a1 aa b7 bc d5 de c3 c8 f9 f2 ef e4 3d 36 2b 20 11 1a 07 0c
        65 6e 73 78 49 42 5f 54 f7 fc e1 ea db d0 cd c6 af a4 b9 b2 83 88 95 9e
        47 4c 51 5a 6b 60 7d 76 1f 14 09 02 33 38 25 2e 8c 87 9a 91 a0 ab b6 bd
        d4 df c2 c9 f8 f3 ee e5 3c 37 2a 21 10 1b 06 0d 64 6f 72 79 48 43 5e 55
        01 0a 17 1c 2d 26 3b 30 59 52 4f 44 75 7e 63 68 b1 ba a7 ac 9d 96 8b 80
        e9 e2 ff f4 c5 ce d3 d8 7a 71 6c 67 56 5d 40 4b 22 29 34 3f 0e 05 18 13
        ca c1 dc d7 e6 ed f0 fb 92 99 84 8f be b5 a8 a3
        """)

    _GM13_data = libutils.str_to_int_array("""
        00 0d 1a 17 34 39 2e 23 68 65 72 7f 5c 51 46 4b d0 dd ca c7 e4 e9 fe f3
        b8 b5 a2 af 8c 81 96 9b bb b6 a1 ac 8f 82 95 98 d3 de c9 c4 e7 ea fd f0
        6b 66 71 7c 5f 52 45 48 03 0e 19 14 37 3a 2d 20 6d 60 77 7a 59 54 43 4e
        05 08 1f 12 31 3c 2b 26 bd b0 a7 aa 89 84 93 9e d5 d8 cf c2 e1 ec fb f6
        d6 db cc c1 e2 ef f8 f5 be b3 a4 a9 8a 87 90 9d 06 0b 1c 11 32 3f 28 25
        6e 63 74 79 5a 57 40 4d da d7 c0 cd ee e3 f4 f9 b2 bf a8 a5 86 8b 9c 91
        0a 07 10 1d 3e 33 24 29 62 6f 78 75 56 5b 4c 41 61 6c 7b 76 55 58 4f 42
        09 04 13 1e 3d 30 27 2a b1 bc ab a6 85 88 9f 92 d9 d4 c3 ce ed e0 f7 fa
        b7 ba ad a0 83 8e 99 94 df d2 c5 c8 eb e6 f1 fc 67 6a 7d 70 53 5e 49 44
        0f 02 15 18 3b 36 21 2c 0c 01 16 1b 38 35 22 2f 64 69 7e 73 50 5d 4a 47
        dc d1 c6 cb e8 e5 f2 ff b4 b9 ae a3 80 8d 9a 97
        """)

    _GM14_data = libutils.str_to_int_array("""
        00 0e 1c 12 38 36 24 2a 70 7e 6c 62 48 46 54 5a e0 ee fc f2 d8 d6 c4 ca
        90 9e 8c 82 a8 a6 b4 ba db d5 c7 c9 e3 ed ff f1 ab a5 b7 b9 93 9d 8f 81
        3b 35 27 29 03 0d 1f 11 4b 45 57 59 73 7d 6f 61 ad a3 b1 bf 95 9b 89 87
        dd d3 c1 cf e5 eb f9 f7 4d 43 51 5f 75 7b 69 67 3d 33 21 2f 05 0b 19 17
        76 78 6a 64 4e 40 52 5c 06 08 1a 14 3e 30 22 2c 96 98 8a 84 ae a0 b2 bc
        e6 e8 fa f4 de d0 c2 cc 41 4f 5d 53 79 77 65 6b 31 3f 2d 23 09 07 15 1b
        a1 af bd b3 99 97 85 8b d1 df cd c3 e9 e7 f5 fb 9a 94 86 88 a2 ac be b0
        ea e4 f6 f8 d2 dc ce c0 7a 74 66 68 42 4c 5e 50 0a 04 16 18 32 3c 2e 20
        ec e2 f0 fe d4 da c8 c6 9c 92 80 8e a4 aa b8 b6 0c 02 10 1e 34 3a 28 26
        7c 72 60 6e 44 4a 58 56 37 39 2b 25 0f 01 13 1d 47 49 5b 55 7f 71 63 6d
        d7 d9 cb c5 ef e1 f3 fd a7 a9 bb b5 9f 91 83 8d
        """)



import io

# the code below disables testing of CompiledSim on systems where there does not appear
# to be the right version of gcc. This is a not an ideal way to check and more work is
# required to more elegantly check compiledsim across multiple architectures.
import subprocess
import unittest

import pyrtl
from pyrtl.corecircuits import _basic_add

try:
    version = subprocess.check_output(["gcc", "--version"])
except OSError as exc:
    msg = "CompiledSimulation testing requires gcc"
    raise unittest.SkipTest(msg) from exc


class TraceWithBasicOpsBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.bitwidth = 3
        self.r = pyrtl.Register(bitwidth=self.bitwidth, name="r")
        self.o = pyrtl.Output(bitwidth=self.bitwidth, name="o")
        self.o <<= self.r

    def check_trace(self, correct_string):
        wtt = pyrtl.working_block().wirevector_subset(pyrtl.Output)
        sim = self.sim(tracer=pyrtl.SimulationTrace(wires_to_track=wtt))
        sim.step_multiple(nsteps=8)
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)

    def test_not_simulation(self):
        self.r.next <<= ~self.r
        self.check_trace("o 07070707\n")

    def test_and_simulation(self):
        self.r.next <<= (~self.r) & pyrtl.Const(6, bitwidth=self.bitwidth)
        self.check_trace("o 06060606\n")

    def test_nand_simulation(self):
        self.r.next <<= self.r.nand(pyrtl.Const(6, bitwidth=self.bitwidth))
        self.check_trace("o 07171717\n")

    def test_or_simulation(self):
        result = self.r | pyrtl.Const(4, bitwidth=self.bitwidth)
        self.r.next <<= result
        self.assertEqual(len(result), self.bitwidth)
        self.check_trace("o 04444444\n")

    def test_xor_simulation(self):
        self.r.next <<= self.r ^ pyrtl.Const(4, bitwidth=self.bitwidth)
        self.check_trace("o 04040404\n")

    def test_plus_simulation(self):
        self.r.next <<= self.r + pyrtl.Const(2, bitwidth=self.bitwidth)
        self.check_trace("o 02460246\n")

    def test_minus_simulation(self):
        self.r.next <<= self.r - pyrtl.Const(1, bitwidth=self.bitwidth)
        self.check_trace("o 07654321\n")

    def test_minus_sim_overflow(self):
        pyrtl.reset_working_block()
        i = pyrtl.Input(8, "i")
        o = pyrtl.Output(name="o")
        o <<= i - 1

        sim = self.sim()
        sim.step({i: 1})
        sim.step({i: 0})
        self.assertEqual(sim.tracer.trace["o"], [0, 0x1FF])

    def test_multiply_simulation(self):
        self.r.next <<= self.r * pyrtl.Const(2, bitwidth=self.bitwidth) + pyrtl.Const(
            1, bitwidth=self.bitwidth
        )
        self.check_trace("o 01377777\n")

    def test_const_nobitwidth_simulation(self):
        self.r.next <<= self.r - pyrtl.Const(1)
        self.check_trace("o 07654321\n")

    def test_const_rawint_simulation(self):
        self.r.next <<= self.r - 1
        self.check_trace("o 07654321\n")

    def test_const_verilogsmall_simulation(self):
        self.r.next <<= self.r - "1'b1"
        self.check_trace("o 07654321\n")

    def test_const_verilogbig_simulation(self):
        self.r.next <<= self.r - "3'b1"
        self.check_trace("o 07654321\n")

    def test_const_veriloghuge_simulation(self):
        self.r.next <<= self.r - "64'b1"
        self.check_trace("o 07654321\n")

    def test_const_veriloghuge2_simulation(self):
        self.r.next <<= self.r + "64'b1"
        self.check_trace("o 01234567\n")

    def test_const_associativity_string_simulation(self):
        self.r.next <<= "64'b1" + self.r
        self.check_trace("o 01234567\n")

    def test_const_associativity_int_simulation(self):
        self.r.next <<= 1 + self.r
        self.check_trace("o 01234567\n")

    def test_bitslice_and_concat_simulation(self):
        left = self.r[0:-1]
        right = pyrtl.Const(1, bitwidth=1)
        self.r.next <<= pyrtl.concat(left, right)
        self.check_trace("o 01377777\n")

    def test_bitslice2_and_concat_simulation(self):
        left = self.r[:-1]
        right = pyrtl.Const(1, bitwidth=1)
        self.r.next <<= pyrtl.concat(left, right)
        self.check_trace("o 01377777\n")

    def test_reg_to_reg_simulation(self):
        self.r2 = pyrtl.Register(bitwidth=self.bitwidth, name="r2")
        self.r.next <<= self.r2
        self.r2.next <<= self.r + pyrtl.Const(2, bitwidth=self.bitwidth)
        self.o2 = pyrtl.Output(bitwidth=self.bitwidth, name="o2")
        self.o2 <<= self.r2
        self.check_trace(" o 00224466\no2 02244660\n")

    def test_less_than_cmp_simulation(self):
        left = self.r[0:-1]
        right = pyrtl.Const(1, bitwidth=1)
        self.r.next <<= left < right
        self.check_trace("o 01010101\n")

    def test_equals_simulation(self):
        left = self.r[0:-1]
        right = pyrtl.Const(0, bitwidth=1)
        self.r.next <<= left == right
        self.check_trace("o 01010101\n")


class PrintTraceBase(unittest.TestCase):
    # note: doesn't include tests for compact=True because all the tests test that
    def setUp(self):
        pyrtl.reset_working_block()
        self.in1 = pyrtl.Input(8, "in1")
        self.in2 = pyrtl.Input(8, "in2")
        self.out = pyrtl.Output(16, "out")

    def test_print_trace_single_dig_notcompact(self):
        self.out <<= pyrtl.probe(self.in1, "in1_probe") + self.in2
        sim = self.sim()
        for i in range(5):
            sim.step({self.in1: i, self.in2: 5 - i})
        correct_outp = (
            "      --- Values in base 10 ---\n"
            "in1       0 1 2 3 4\n"
            "in1_probe 0 1 2 3 4\n"
            "in2       5 4 3 2 1\n"
            "out       5 5 5 5 5\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_outp)

    def test_print_trace_base2(self):
        self.out <<= pyrtl.probe(self.in1, "in1_probe") + self.in2
        sim = self.sim()
        for i in range(5):
            sim.step({self.in1: 4 * i, self.in2: 4 * (5 - i)})
        correct_outp = (
            "      --- Values in base 2 ---\n"
            "in1           0   100  1000  1100 10000\n"
            "in1_probe     0   100  1000  1100 10000\n"
            "in2       10100 10000  1100  1000   100\n"
            "out       10100 10100 10100 10100 10100\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output, base=2)
        self.assertEqual(output.getvalue(), correct_outp)

    def test_print_trace_base8(self):
        self.out <<= pyrtl.probe(self.in1, "in1_probe") + self.in2
        sim = self.sim()
        for i in range(5):
            sim.step({self.in1: 6 * i, self.in2: 6 * (5 - i)})
        correct_outp = (
            "      --- Values in base 8 ---\n"
            "in1        0  6 14 22 30\n"
            "in1_probe  0  6 14 22 30\n"
            "in2       36 30 22 14  6\n"
            "out       36 36 36 36 36\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output, base=8)
        self.assertEqual(output.getvalue(), correct_outp)

    def test_print_trace_base16(self):
        self.out <<= pyrtl.probe(self.in1, "in1_probe") * self.in2
        sim = self.sim()
        for i in range(5):
            sim.step({self.in1: 9 * i, self.in2: 9 * (5 - i)})
        correct_outp = (
            "      --- Values in base 16 ---\n"
            "in1         0   9  12  1b  24\n"
            "in1_probe   0   9  12  1b  24\n"
            "in2        2d  24  1b  12   9\n"
            "out         0 144 1e6 1e6 144\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output, base=16)
        self.assertEqual(output.getvalue(), correct_outp)


class SimWithSpecialWiresBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_reg_directly_before_reg(self):
        pass

    def test_weird_wire_names(self):
        """Some simulations need to be careful when handling special names (eg Fastsim
        June 2016)
        """
        i = pyrtl.Input(8, '"182&!!!\n')
        o = pyrtl.Output(8, "*^*)#*$'*")
        o2 = pyrtl.Output(8, "test@+")
        w = pyrtl.WireVector(8, "[][[-=--09888")
        r = pyrtl.Register(8, "&@#)^#@^&(asdfkhafkjh")

        w <<= i
        r.next <<= i
        o <<= w
        o2 <<= r

        trace = pyrtl.SimulationTrace()
        sim = self.sim(tracer=trace)

        sim.step({i: 28})
        self.assertEqual(trace.trace[o.name], [28])
        sim.step({i: 233})
        self.assertEqual(trace.trace[o2.name], [0, 28])

    def test_fastsim_wire_names(self):
        """Testing ability to use wire names instead of wires in input"""
        in1 = pyrtl.Input(8, "in1")
        in2 = pyrtl.Input(8, "in2")
        in3 = pyrtl.Input(8, "in3")
        truth = pyrtl.Const(1, 1)
        out1 = pyrtl.Output(16, "out2")
        out2 = pyrtl.Output(16, "out3")
        out1 <<= in1 + in2
        out2 <<= in3 | truth
        sim = self.sim()
        for i in range(10):
            sim.step({"in1": 2 * i, "in2": 3 * i, "in3": 40 - 2 * i})
        correct_outp = (
            " --- Values in base 10 ---\n"
            "in1   0  2  4  6  8 10 12 14 16 18\n"
            "in2   0  3  6  9 12 15 18 21 24 27\n"
            "in3  40 38 36 34 32 30 28 26 24 22\n"
            "out2  0  5 10 15 20 25 30 35 40 45\n"
            "out3 41 39 37 35 33 31 29 27 25 23\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_outp)


class SimInputValidationBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_input_out_of_bitwidth(self):
        counter = pyrtl.Register(bitwidth=3, name="counter")
        i = pyrtl.Input(bitwidth=2, name="i")
        counter.next <<= counter + i

        sim = self.sim()
        for cycle in range(4):
            sim.step({i: cycle})
        with self.assertRaises(pyrtl.PyrtlError):
            sim.step({i: 5})


class SimStepMultipleBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        in1 = pyrtl.Input(4, "in1")
        in2 = pyrtl.Input(4, "in2")
        out1 = pyrtl.Output(4, "out1")
        out1 <<= (in1 ^ in2) + pyrtl.Const(1)
        out2 = pyrtl.Output(4, "out2")
        out2 <<= in1 | in2
        self.inputs = {
            "in1": [0, 1, 3, 15, 14],
            # When a string, assumes each input is a single digit integer.
            "in2": "66666",
        }

    def test_step_multiple_nsteps_no_inputs(self):
        pyrtl.reset_working_block()
        a = pyrtl.Register(8)
        b = pyrtl.Output(8, "b")
        a.next <<= a + 1
        b <<= a

        sim = self.sim()
        sim.step_multiple(nsteps=5)

        correct_output = "--- Values in base 10 ---\nb 0 1 2 3 4\n"
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_output)

    def test_step_multiple_nsteps_gt_ninputs(self):
        sim = self.sim()

        with self.assertRaises(pyrtl.PyrtlError) as error:
            sim.step_multiple(self.inputs, nsteps=6)
        self.assertEqual(
            str(error.exception),
            "nsteps is specified but is greater than the "
            "number of values supplied for each input",
        )

    def test_step_multiple_no_inputs(self):
        sim = self.sim()

        with self.assertRaises(pyrtl.PyrtlError) as error:
            sim.step_multiple()
        self.assertEqual(
            str(error.exception),
            "need to supply either input values or a number of steps to simulate",
        )

    def test_step_multiple_bad_nsteps1(self):
        sim = self.sim()

        with self.assertRaises(pyrtl.PyrtlError) as error:
            sim.step_multiple({"in1": [], "in2": []})
        self.assertEqual(str(error.exception), "must simulate at least one step")

    def test_step_multiple_bad_nsteps2(self):
        sim = self.sim()

        with self.assertRaises(pyrtl.PyrtlError) as error:
            sim.step_multiple(self.inputs, nsteps=-1)
        self.assertEqual(str(error.exception), "must simulate at least one step")

    def test_step_multiple_no_values_for_each_step_of_input(self):
        sim = self.sim()

        with self.assertRaises(pyrtl.PyrtlError) as error:
            sim.step_multiple({"in1": [0, 1, 3], "in2": [1]})
        self.assertEqual(
            str(error.exception),
            "must supply a value for each provided wire for each step of simulation",
        )

    def test_step_multiple_no_values_for_each_step_of_given_outputs(self):
        sim = self.sim()

        with self.assertRaises(pyrtl.PyrtlError) as error:
            sim.step_multiple(
                self.inputs, {"out1": [7, 8, 6, 10, 9], "out2": [6, 7, 7, 15]}
            )
        self.assertEqual(
            str(error.exception),
            "any expected outputs must have a supplied value each step of simulation",
        )

    def test_step_multiple_no_expected_check(self):
        sim = self.sim()

        sim.step_multiple(self.inputs)

        correct_output = (
            " --- Values in base 10 ---\n"
            "in1   0  1  3 15 14\n"
            "in2   6  6  6  6  6\n"
            "out1  7  8  6 10  9\n"
            "out2  6  7  7 15 14\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_output)

    def test_step_multiple_no_errors(self):
        sim = self.sim()

        expected = {
            "out1": [7, 8, 6, 10, 9],
            "out2": [6, 7, 7, 15, 14],
        }
        sim.step_multiple(self.inputs, expected)

        correct_output = (
            " --- Values in base 10 ---\n"
            "in1   0  1  3 15 14\n"
            "in2   6  6  6  6  6\n"
            "out1  7  8  6 10  9\n"
            "out2  6  7  7 15 14\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_output)

    def test_step_multiple_dont_care_expected(self):
        sim = self.sim()

        expected = {
            "out1": [7, "?", 6, 10],
            "out2": "6?7?",
        }
        sim.step_multiple(self.inputs, expected, nsteps=4)

        correct_output = (
            " --- Values in base 10 ---\n"
            "in1   0  1  3 15\n"
            "in2   6  6  6  6\n"
            "out1  7  8  6 10\n"
            "out2  6  7  7 15\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_output)

    def test_step_multiple_no_errors_nsteps_specified(self):
        sim = self.sim()

        expected = {
            "out1": [7, 8, 6, 10, 9],
            "out2": [6, 7, 7, 15, 14],
        }
        sim.step_multiple(self.inputs, expected, nsteps=3)

        correct_output = (
            " --- Values in base 10 ---\n"
            "in1  0 1 3\n"
            "in2  6 6 6\n"
            "out1 7 8 6\n"
            "out2 6 7 7\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_output)

    def test_step_multiple_many_errors_report_only_first(self):
        sim = self.sim()

        expected = {
            "out1": [7, 9, 4, 10, 9],
            "out2": [6, 2, 7, 8, 14],
        }
        output = io.StringIO()
        sim.step_multiple(
            self.inputs, expected, file=output, stop_after_first_error=True
        )

        # Test the output about unexpected values
        correct_output = (
            "Unexpected output (stopped after step with first error):\n"
            " step       name expected   actual\n"
            "    1       out1        9        8\n"
            "    1       out2        2        7\n"
        )
        self.assertEqual(output.getvalue(), correct_output)

        # Test that the trace stopped after the step with the first error
        correct_output = (
            " --- Values in base 10 ---\nin1  0 1\nin2  6 6\nout1 7 8\nout2 6 7\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_output)

    def test_step_multiple_many_errors_report_all(self):
        sim = self.sim()

        expected = {
            "out1": [7, 9, 4, 10, 9],
            "out2": [6, 2, 7, 8, 14],
        }
        output = io.StringIO()
        sim.step_multiple(self.inputs, expected, file=output)

        # Test the output about unexpected values
        correct_output = (
            "Unexpected output on one or more steps:\n"
            " step       name expected   actual\n"
            "    1       out1        9        8\n"
            "    1       out2        2        7\n"
            "    2       out1        4        6\n"
            "    3       out2        8       15\n"
        )
        self.assertEqual(output.getvalue(), correct_output)

        # Test that the trace still produced all the steps
        correct_output = (
            " --- Values in base 10 ---\n"
            "in1   0  1  3 15 14\n"
            "in2   6  6  6  6  6\n"
            "out1  7  8  6 10  9\n"
            "out2  6  7  7 15 14\n"
        )
        output = io.StringIO()
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_output)


class TraceWithAdderBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        bitwidth = 3
        self.r = pyrtl.Register(bitwidth=bitwidth, name="r")
        self.result = _basic_add(self.r, pyrtl.Const(1).zero_extended(bitwidth))
        self.r.next <<= self.result
        self.o = pyrtl.Output(bitwidth=bitwidth, name="o")
        self.o <<= self.r

    def test_adder_simulation(self):
        sim_trace = pyrtl.SimulationTrace(wires_to_track=[self.o])

        # build the actual simulation environment
        sim = self.sim(tracer=sim_trace)
        sim.step_multiple(nsteps=15)

        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        file = io.StringIO()
        sim.tracer.render_trace(file=file)  # want to make sure the code at least runs
        self.assertEqual(output.getvalue(), "o 012345670123456\n")


class SimulationVCDWithAdderBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        bitwidth = 3
        self.r = pyrtl.Register(bitwidth=bitwidth, name="r")
        self.result = _basic_add(self.r, pyrtl.Const(1).zero_extended(bitwidth))
        self.r.next <<= self.result
        self.o = pyrtl.Output(bitwidth=bitwidth, name="o")
        self.o <<= self.r

    VCD_OUTPUT = """$timescale 1ns $end
$scope module logic $end
$var wire 3 o o $end
$upscope $end
$enddefinitions $end
$dumpvars
b0 o
$end
#0
b0 o

#10
b1 o

#20
b10 o

#30
b11 o

#40
b100 o

#50
b101 o

#60
b110 o

#70
b111 o

#80
b0 o

#90
b1 o

#100
b10 o

#110
b11 o

#120
b100 o

#130
b101 o

#140
b110 o

#150
"""

    def test_vcd_output(self):
        sim_trace = pyrtl.SimulationTrace(wires_to_track=[self.o])

        # build the actual simulation environment
        sim = self.sim(tracer=sim_trace)
        sim.step_multiple(nsteps=15)

        test_output = io.StringIO()
        sim.tracer.print_vcd(test_output)
        self.assertEqual(self.VCD_OUTPUT, test_output.getvalue())


class SimTraceWithMuxBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        bitwidth = 3
        self.a = pyrtl.Input(bitwidth=bitwidth)
        self.b = pyrtl.Input(bitwidth=bitwidth)
        self.sel = pyrtl.Input(bitwidth=1)
        self.muxout = pyrtl.Output(bitwidth=bitwidth, name="muxout")
        self.muxout <<= pyrtl.mux(self.sel, self.a, self.b)

        # build the actual simulation environment
        self.sim = self.sim()

    def test_adder_simulation(self):
        input_signals = {
            0: {self.a: 0, self.b: 1, self.sel: 1},
            1: {self.a: 0, self.b: 2, self.sel: 1},
            2: {self.a: 0, self.b: 0, self.sel: 1},
            3: {self.a: 1, self.b: 1, self.sel: 0},
            4: {self.a: 2, self.b: 1, self.sel: 0},
            5: {self.a: 0, self.b: 1, self.sel: 0},
        }
        for i in range(6):
            self.sim.step(input_signals[i])

        output = io.StringIO()
        self.sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), "muxout 120120\n")


class MemBlockBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.bitwidth = 3
        self.addrwidth = 3
        self.output1 = pyrtl.Output(self.bitwidth, "o1")
        self.output2 = pyrtl.Output(self.bitwidth, "o2")
        self.read_addr1 = pyrtl.Input(self.addrwidth)
        self.read_addr2 = pyrtl.Input(self.addrwidth)
        self.write_addr = pyrtl.Input(self.addrwidth)
        self.write_data = pyrtl.Input(self.bitwidth)
        self.mem1 = pyrtl.MemBlock(
            bitwidth=self.bitwidth, addrwidth=self.addrwidth, name="mem1"
        )
        self.mem2 = pyrtl.MemBlock(
            bitwidth=self.bitwidth, addrwidth=self.addrwidth, name="mem2"
        )
        self.output1 <<= self.mem1[self.read_addr1]
        self.output2 <<= self.mem1[self.read_addr2]
        self.mem1[self.write_addr] <<= self.write_data

    def test_simple_memblock(self):
        sim = self.sim()

        input_signals = [
            [0, 1, 4, 5],
            [4, 1, 0, 5],
            [0, 4, 1, 6],
            [1, 1, 0, 0],
            [6, 0, 6, 7],
        ]
        for signals in input_signals:
            sim.step(
                {
                    self.read_addr1: signals[0],
                    self.read_addr2: signals[1],
                    self.write_addr: signals[2],
                    self.write_data: signals[3],
                }
            )

        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), "o1 05560\no2 00560\n")

    def test_simple2_memblock(self):
        sim = self.sim()
        input_signals = [
            {
                self.read_addr1: 0,
                self.read_addr2: 1,
                self.write_addr: 0,
                self.write_data: 0x7,
            },
            {
                self.read_addr1: 1,
                self.read_addr2: 2,
                self.write_addr: 1,
                self.write_data: 0x6,
            },
            {
                self.read_addr1: 0,
                self.read_addr2: 0,
                self.write_addr: 2,
                self.write_data: 0x5,
            },
            {
                self.read_addr1: 0,
                self.read_addr2: 1,
                self.write_addr: 0,
                self.write_data: 0x4,
            },
            {
                self.read_addr1: 1,
                self.read_addr2: 0,
                self.write_addr: 1,
                self.write_data: 0x3,
            },
            {
                self.read_addr1: 2,
                self.read_addr2: 2,
                self.write_addr: 2,
                self.write_data: 0x2,
            },
            {
                self.read_addr1: 1,
                self.read_addr2: 2,
                self.write_addr: 0,
                self.write_data: 0x1,
            },
            {
                self.read_addr1: 0,
                self.read_addr2: 1,
                self.write_addr: 1,
                self.write_data: 0x0,
            },
            {
                self.read_addr1: 1,
                self.read_addr2: 0,
                self.write_addr: 2,
                self.write_data: 0x7,
            },
            {
                self.read_addr1: 2,
                self.read_addr2: 1,
                self.write_addr: 0,
                self.write_data: 0x6,
            },
        ]

        for signal in input_signals:
            sim.step(signal)

        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), "o1 0077653107\no2 0076452310\n")

    def test_synth_simple_memblock(self):
        pyrtl.synthesize()
        pyrtl.optimize()
        sim = self.sim()
        input_signals = [
            [0, 1, 4, 5],
            [4, 1, 0, 5],
            [0, 4, 1, 6],
            [1, 1, 0, 0],
            [6, 0, 6, 7],
        ]
        for signals in input_signals:
            sim.step(
                {
                    self.read_addr1: signals[0],
                    self.read_addr2: signals[1],
                    self.write_addr: signals[2],
                    self.write_data: signals[3],
                }
            )

        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), "o1 05560\no2 00560\n")

    def test_mem_val_map(self):
        read_addr3 = pyrtl.Input(self.addrwidth)
        self.output3 = pyrtl.Output(self.bitwidth, "o3")
        self.output3 <<= self.mem2[read_addr3]
        mem_val_map = {
            self.mem1: {0: 0, 1: 1, 2: 2, 3: 3},
            self.mem2: {0: 4, 1: 5, 2: 6, 3: 7},
        }
        sim = self.sim(memory_value_map=mem_val_map)
        # put new entries in
        for i in range(2):
            sim.step(
                {
                    self.read_addr1: 4 + i,  # d.c.
                    self.read_addr2: 4 + i,  # d.c.
                    read_addr3: 2,
                    self.write_addr: 4
                    + i,  # put a 4 and a 5 in the 4th and 5th addr of mem1
                    self.write_data: 4 + i,
                }
            )
        # modify existing entries
        for i in range(2):
            sim.step(
                {
                    self.read_addr1: 1 + i,  # d.c.
                    self.read_addr2: 1 + i,  # d.c.
                    read_addr3: 2,
                    self.write_addr: 1
                    + i,  # put a 2 and a 3 in the 1st and 2nd addr of mem1
                    self.write_data: 2 + i,
                }
            )
        # check consistency of memory_value_map assignment, insertion, and modification
        self.assertEqual(
            sim.inspect_mem(self.mem1), {0: 0, 1: 2, 2: 3, 3: 3, 4: 4, 5: 5}
        )
        self.assertEqual(sim.tracer.memory_value_map, mem_val_map)

    def test_mem_val_map_defaults(self):
        read_addr3 = pyrtl.Input(self.addrwidth)
        self.output3 = pyrtl.Output(self.bitwidth, "o3")
        self.output3 <<= self.mem2[read_addr3]
        mem_val_map = {self.mem1: {0: 0, 1: 1}, self.mem2: {0: 4, 1: 5}}
        sim = self.sim(memory_value_map=mem_val_map)
        for i in range(2, 8):
            sim.step(
                {
                    self.read_addr1: i,
                    self.read_addr2: 8 - i + 1,
                    read_addr3: i,
                    self.write_addr: 0,
                    self.write_data: 0,
                }
            )
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), "o1 000000\no2 000000\no3 000000\n")
        self.assertEqual(sim.tracer.memory_value_map, mem_val_map)

    def test_mem_val_map_empty_mapping(self):
        read_addr3 = pyrtl.Input(self.addrwidth)
        self.output3 = pyrtl.Output(self.bitwidth, "o3")
        self.output3 <<= self.mem2[read_addr3]
        mem_val_map = {self.mem1: {0: 0, 1: 1}, self.mem2: {}}
        sim = self.sim(memory_value_map=mem_val_map)
        for i in range(2, 8):
            sim.step(
                {
                    self.read_addr1: i,
                    self.read_addr2: 8 - i + 1,
                    read_addr3: i,
                    self.write_addr: 0,
                    self.write_data: 0,
                }
            )
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), "o1 000000\no2 000000\no3 000000\n")
        self.assertEqual(sim.tracer.memory_value_map, mem_val_map)


class MemBlockLargeBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.bitwidth = 68
        self.addrwidth = 32
        self.output1 = pyrtl.Output(self.bitwidth, "o1")
        self.output2 = pyrtl.Output(self.bitwidth, "o2")
        self.read_addr1 = pyrtl.Input(self.addrwidth)
        self.read_addr2 = pyrtl.Input(self.addrwidth)
        self.write_addr = pyrtl.Input(self.addrwidth)
        self.write_data = pyrtl.Input(self.bitwidth)
        self.mem = pyrtl.MemBlock(
            bitwidth=self.bitwidth, addrwidth=self.addrwidth, name="mem"
        )
        self.output1 <<= self.mem[self.read_addr1]
        self.output2 <<= self.mem[self.read_addr2]
        self.mem[self.write_addr] <<= self.write_data

    def test_mem_blocks_very_large(self):
        """Tests support of very large memories (i.e. address width > 30 bits), and that
        limbs are handled appropriately for bitwidths > 64
        """
        sim = self.sim()

        write_data = 0x20000000040000012  # 68 bits
        input_signals = [
            [0, 1, 0xFFFFFFFF, write_data],
            [0xFFFFFFFF, 1, 0, write_data],
            [0, 0xFFFFFFFF, 0xF0000001, 6],
            [0xF0000001, 0xF0000001, 0, 0],
            [6, 0, 6, 7],
        ]
        for signals in input_signals:
            sim.step(
                {
                    self.read_addr1: signals[0],
                    self.read_addr2: signals[1],
                    self.write_addr: signals[2],
                    self.write_data: signals[3],
                }
            )

        output = io.StringIO()
        correct_outp = (
            "--- Values in base 10 ---\n"
            f"o1                    0 {write_data} {write_data}                    6   "
            "                 0\n"
            f"o2                    0                    0 {write_data}                "
            "    6                    0\n"
        )
        sim.tracer.print_trace(output)
        self.assertEqual(output.getvalue(), correct_outp)


class RegisterDefaultsBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.i = pyrtl.Input(bitwidth=3)
        self.r1 = pyrtl.Register(name="r1", bitwidth=3)
        self.r2 = pyrtl.Register(name="r2", bitwidth=3, reset_value=7)
        self.o = pyrtl.Output(name="o", bitwidth=3)
        self.r1.next <<= self.i
        self.r2.next <<= self.r1
        self.o <<= self.r2

    def check_trace(self, correct_string, **kwargs):
        wtt = pyrtl.working_block().wirevector_subset(pyrtl.Output)
        sim = self.sim(tracer=pyrtl.SimulationTrace(wires_to_track=wtt), **kwargs)
        for i in range(8):
            sim.step({self.i: i})
        output = io.StringIO()
        sim.tracer.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), correct_string)
        return sim.tracer

    def test_reset_value(self):
        self.check_trace("o 70012345\n")

    def test_register_map_overrides_reset_value(self):
        register_value_map = {self.r1: 6, self.r2: 3}
        tracer = self.check_trace("o 36012345\n", register_value_map=register_value_map)
        self.assertEqual(tracer.register_value_map, register_value_map)

    def test_partial_map(self):
        register_value_map = {self.r1: 6}
        tracer = self.check_trace("o 76012345\n", register_value_map=register_value_map)
        self.assertEqual(tracer.register_value_map, register_value_map)

    def test_partial_map_overrides_default(self):
        register_value_map = {self.r1: 6}
        tracer = self.check_trace(
            "o 76012345\n", default_value=5, register_value_map=register_value_map
        )
        self.assertEqual(tracer.default_value, 5)
        self.assertEqual(tracer.register_value_map, register_value_map)

    def test_default_used_for_non_reset_value(self):
        tracer = self.check_trace("o 75012345\n", default_value=5)
        self.assertEqual(tracer.default_value, 5)


class RomBlockSimBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def generate_expected_output(self, data_tuples, length):
        """dataTuple is in a series of tuples in (name, function) format.

        the function takes in a single argument, length. length is the number of steps
        in the length
        """

        out_string = ""
        for tuple in data_tuples:
            out_string = out_string + tuple[0] + " "
            for time in range(0, length - 1):
                out_string += str(tuple[1](time))
            out_string += "\n"
        return out_string

    def compareIO(self, sim_trace_a, expected_output):
        output = io.StringIO()
        sim_trace_a.print_trace(output, compact=True)
        self.assertEqual(output.getvalue(), expected_output)

    def rom_block_test_helper(self, bitwidth):
        def rom_data_function(add):
            return int((add + 5) / 2)

        pyrtl.reset_working_block()
        self.bitwidth = bitwidth
        self.addrwidth = 4
        self.output1 = pyrtl.Output(self.bitwidth, "o1")
        self.output2 = pyrtl.Output(self.bitwidth, "o2")
        self.read_addr1 = pyrtl.Input(self.addrwidth)
        self.read_addr2 = pyrtl.Input(self.addrwidth)
        self.rom = pyrtl.RomBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name="rom",
            romdata=rom_data_function,
        )
        self.output1 <<= self.rom[self.read_addr1]
        self.output2 <<= self.rom[self.read_addr2]
        # build the actual simulation environment
        self.sim = self.sim()

        input_signals = {}
        for i in range(0, 5):
            input_signals[i] = {self.read_addr1: i, self.read_addr2: 2 * i}
            self.sim.step(input_signals[i])

        exp_out = self.generate_expected_output(
            (
                ("o1", lambda x: rom_data_function(x)),
                ("o2", lambda x: rom_data_function(2 * x)),
            ),
            6,
        )
        self.compareIO(self.sim.tracer, exp_out)

    def test_function_rom_block_bitwidth_4(self):
        self.rom_block_test_helper(4)

    def test_function_rom_block_bitwidth_12(self):
        self.rom_block_test_helper(12)

    def test_function_rom_block_bitwidth_32(self):
        self.rom_block_test_helper(32)

    def test_function_rom_block_bitwidth_64(self):
        self.rom_block_test_helper(64)

    def rom_block_with_optimization_helper(self, bitwidth):
        def rom_data_function(add):
            return int((add + 5) / 2)

        pyrtl.reset_working_block()
        self.bitwidth = bitwidth
        self.addrwidth = 4
        self.output1 = pyrtl.Output(self.bitwidth, "o1")
        self.output2 = pyrtl.Output(self.bitwidth, "o2")

        self.read_addr1 = pyrtl.Input(self.addrwidth)
        self.read_addr2 = pyrtl.Input(self.addrwidth)
        self.rom = pyrtl.RomBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name="rom",
            romdata=rom_data_function,
        )
        self.output1 <<= self.rom[self.read_addr1]
        self.output2 <<= self.rom[self.read_addr2]

        pyrtl.synthesize()
        pyrtl.optimize()
        # build the actual simulation environment
        self.sim = self.sim()

        input_signals = {}
        for i in range(0, 5):
            input_signals[i] = {self.read_addr1: i, self.read_addr2: 2 * i}
            input_signals[i] = {self.read_addr1: i, self.read_addr2: 2 * i}
            self.sim.step(input_signals[i])

        exp_out = self.generate_expected_output(
            (
                ("o1", lambda x: rom_data_function(x)),
                ("o2", lambda x: rom_data_function(2 * x)),
            ),
            6,
        )
        self.compareIO(self.sim.tracer, exp_out)

    def test_function_rom_block_with_optimization_bitwidth_4(self):
        self.rom_block_with_optimization_helper(4)

    def test_function_rom_block_with_optimization_bitwidth_12(self):
        self.rom_block_with_optimization_helper(12)

    def test_function_rom_block_with_optimization_bitwidth_32(self):
        self.rom_block_with_optimization_helper(32)

    def test_function_rom_block_with_optimization_bitwidth_64(self):
        self.rom_block_with_optimization_helper(64)

    def test_rom_out_of_range_error(self):
        rom_data_array = [15, 13, 11, 9, 7, 5, 3]
        rom1 = pyrtl.RomBlock(bitwidth=4, addrwidth=3, romdata=rom_data_array)
        rom_add_1 = pyrtl.Input(3, "rom_in")
        rom_out_1 = pyrtl.Output(4, "rom_out_1")
        rom_out_1 <<= rom1[rom_add_1]

        with self.assertRaises(pyrtl.PyrtlError):
            self.sim()

    def test_rom_rom_block_in_memory_value_map_error(self):
        rom_data_array = [6]
        rom1 = pyrtl.RomBlock(bitwidth=4, addrwidth=1, romdata=rom_data_array)
        rom_addr = pyrtl.Input(1, "rom_addr")
        rom_out = pyrtl.Output(4, "rom_out")
        rom_out <<= rom1[rom_addr]
        mem_val_map = {rom1: {0: 0}}

        with self.assertRaises(pyrtl.PyrtlError) as error:
            self.sim(memory_value_map=mem_val_map)
        self.assertIn("memory_value_map", str(error.exception))

    def test_rom_val_map(self):
        def rom_data_function(add):
            return int((add + 5) / 2)

        self.bitwidth = 4
        self.addrwidth = 4
        self.rom1 = pyrtl.RomBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name="rom1",
            romdata=rom_data_function,
        )
        self.rom2 = pyrtl.RomBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name="rom2",
            romdata=rom_data_function,
        )
        mem_val_map = {
            self.rom1: {0: 0, 1: 1, 2: 2, 3: 3},
            self.rom2: {0: 4, 1: 5, 2: 6, 3: 7},
        }

        with self.assertRaises(pyrtl.PyrtlError):
            _ = self.sim(memory_value_map=mem_val_map)

    def test_negative_memory_value_map(self):
        mem = pyrtl.MemBlock(addrwidth=3, bitwidth=3)
        counter = pyrtl.Register(bitwidth=mem.addrwidth)
        counter.next <<= counter + 1
        read_data = pyrtl.Output(name="read_data")
        read_data <<= mem[counter]
        memory_values = [-4, -3, -2, -1, 0, 1, 2, 3]
        memory_value_map = dict(enumerate(memory_values))
        sim = self.sim(memory_value_map={mem: memory_value_map})
        for i in range(2**mem.addrwidth):
            sim.step()
            actual_read_data = pyrtl.val_to_signed_integer(
                sim.inspect("read_data"), bitwidth=mem.bitwidth
            )
            self.assertEqual(actual_read_data, memory_value_map[i])

    def test_unsigned_memory_value_map(self):
        mem = pyrtl.MemBlock(addrwidth=3, bitwidth=3)
        counter = pyrtl.Register(bitwidth=mem.addrwidth)
        counter.next <<= counter + 1
        read_data = pyrtl.Output(name="read_data")
        read_data <<= mem[counter]
        memory_values = [0, 1, 2, 3, 4, 5, 6, 7]
        memory_value_map = dict(enumerate(memory_values))
        sim = self.sim(memory_value_map={mem: memory_value_map})
        for i in range(2**mem.addrwidth):
            sim.step()
            actual_read_data = sim.inspect("read_data")
            self.assertEqual(actual_read_data, memory_value_map[i])


class InspectBase(unittest.TestCase):
    """Unittests for sim.inspect_mem."""

    def setUp(self):
        pyrtl.reset_working_block()

    def test_inspect_mem(self):
        a = pyrtl.Input(8, "a")
        b = pyrtl.Input(8, "b")
        mem = pyrtl.MemBlock(8, 8, "mem")
        mem[b] <<= a
        sim = self.sim()
        self.assertEqual(sim.inspect_mem(mem), {})
        sim.step({a: 3, b: 23})
        self.assertEqual(sim.inspect_mem(mem), {23: 3})


class TraceErrorBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_empty_trace(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.SimulationTrace()

    def test_empty_trace_after_untraceable_removed(self):
        r = pyrtl.Register(2, "r")
        r.next <<= r + 1
        sim = self.sim()
        sim.step_multiple(provided_inputs={}, nsteps=10)
        with self.assertRaises(pyrtl.PyrtlError) as ex:
            sim.tracer.render_trace()
        self.assertEqual(
            str(ex.exception),
            "Empty trace list. This may have occurred because "
            "untraceable wires were removed prior to simulation, "
            "if a CompiledSimulation was used.",
        )

    def test_invalid_base(self):
        self.in1 = pyrtl.Input(8, "in1")
        self.out = pyrtl.Output(8, "out")
        self.out <<= self.in1
        sim = self.sim()
        for i in range(5):
            sim.step({self.in1: i})
        with self.assertRaises(pyrtl.PyrtlError):
            sim.tracer.print_trace(base=4)


def make_unittests():
    """Generates separate unittests for each of the simulators."""
    g = globals()
    unittests = {}
    base_tests = {
        name: v
        for name, v in g.items()
        if isinstance(v, type) and issubclass(v, unittest.TestCase)
    }
    for name, v in base_tests.items():
        del g[name]
        if name[-4:].lower() == "base":
            name = name[:-4]
        else:
            # Add Base to the end of your unittest name to tell this that you actually
            # made the unittest as such
            msg = (
                "You should be making unittests that are compatible with both "
                "FastSimulation and Simulation"
            )
            raise Exception(msg)
        for sim in sims:
            unit_name = "Test" + name + sim.__name__
            unittests[unit_name] = type(unit_name, (v,), {"sim": sim})
    g.update(unittests)


sims = (pyrtl.CompiledSimulation,)
make_unittests()


if __name__ == "__main__":
    unittest.main()



import doctest
import unittest

import pyrtl


class TestDocTests(unittest.TestCase):
    """Test documentation examples."""

    def test_doctests(self):
        failures, tests = doctest.testmod(
            m=pyrtl.wire, optionflags=doctest.IGNORE_EXCEPTION_DETAIL
        )
        self.assertGreater(tests, 0)
        self.assertEqual(failures, 0)


class TestWireVector(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_wirevector_length(self):
        x = pyrtl.WireVector(1)
        y = pyrtl.WireVector(13)
        self.assertEqual(len(x), 1)
        self.assertEqual(len(y), 13)

    def test_basic_assignment(self):
        x = pyrtl.WireVector(1)
        y = pyrtl.WireVector(1)
        y <<= x

    def test_unset_bitwidth(self):
        x = pyrtl.Input()
        y = pyrtl.WireVector(1)
        with self.assertRaises(pyrtl.PyrtlError):
            y <<= x

    def test_assign_to_value(self):
        x = pyrtl.WireVector(1)
        y = 1337
        with self.assertRaises(TypeError):
            y <<= x

    def test_zero_extend(self):
        pass

    def test_sign_extend(self):
        pass

    def test_truncating(self):
        pass

    def test_rename(self):
        block = pyrtl.working_block()
        w = pyrtl.WireVector(1, "test1")
        self.assertIn("test1", block.wirevector_by_name)
        self.assertIn(w, block.wirevector_set)
        w.name = "testJohn"
        self.assertNotIn("test1", block.wirevector_by_name)
        self.assertIn("testJohn", block.wirevector_by_name)
        self.assertIn(w, block.wirevector_set)


class TestWireVectorNames(unittest.TestCase):
    def is_valid_str(self, s):
        return pyrtl.wire.next_tempvar_name(s) == s

    def test_invalid_name(self):
        self.assertFalse(self.is_valid_str(""))
        with self.assertRaises(pyrtl.PyrtlError):
            self.is_valid_str("clock")

    def test_valid_names(self):
        self.assertTrue(self.is_valid_str("xxx"))
        self.assertTrue(self.is_valid_str("h"))
        self.assertTrue(self.is_valid_str(" "))
        self.assertTrue(self.is_valid_str("#$)(*&#@_+!#)("))

    def check_name_setter(self, ns):
        """test name setter in wire.py.

        Keyword arguments:

        ns -- new name of WireVector to be set
        """
        test = pyrtl.WireVector(1, "test")
        test.name = ns
        self.assertTrue(test.name == ns)

    def test_invalid_name_setter(self):
        """test invalid name data types and expect PyrtlError."""
        with self.assertRaises(pyrtl.PyrtlError):
            self.check_name_setter(24)
        with self.assertRaises(pyrtl.PyrtlError):
            self.check_name_setter(True)
        with self.assertRaises(pyrtl.PyrtlError):
            self.check_name_setter(3.14)

    def test_valid_name_setter(self):
        """test string names and expect no error."""
        self.check_name_setter("24")
        self.check_name_setter(str(24))
        self.check_name_setter("twenty_four")


class TestWireVectorFail(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_undef_wirevector_length(self):
        x = pyrtl.WireVector()
        with self.assertRaises(pyrtl.PyrtlError):
            _ = len(x)

    def test_bad_bitwidth(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.WireVector(bitwidth="happy")
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.WireVector(bitwidth=-1)
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.WireVector(bitwidth=0)

        y = pyrtl.WireVector(1)
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.WireVector(y)

    def test_no_immed_operators(self):
        x = pyrtl.WireVector(bitwidth=3)
        with self.assertRaises(pyrtl.PyrtlError):
            x &= 2
        with self.assertRaises(pyrtl.PyrtlError):
            x ^= 2
        with self.assertRaises(pyrtl.PyrtlError):
            x += 2
        with self.assertRaises(pyrtl.PyrtlError):
            x -= 2
        with self.assertRaises(pyrtl.PyrtlError):
            x *= 2

    def test_sign_and_zero_extend_only_increase_bitwidth(self):
        x = pyrtl.WireVector(bitwidth=3)
        with self.assertRaises(pyrtl.PyrtlError):
            x.zero_extended(2)

    def test_truncate_only_reduces_bitwidth(self):
        x = pyrtl.WireVector(bitwidth=3)
        with self.assertRaises(pyrtl.PyrtlError):
            x.truncate(5)

    def test_truncate_only_takes_integers(self):
        x = pyrtl.WireVector(bitwidth=3)
        y = pyrtl.WireVector(bitwidth=3)
        with self.assertRaises(pyrtl.PyrtlError):
            x.truncate(y)


class TestWireVectorSlicing(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def invalid_empty_slice(self, bitwidth, slice):
        w = pyrtl.Input(bitwidth)
        with self.assertRaises(pyrtl.PyrtlError):
            _ = w[slice]

    def invalid_slice_index(self, bitwidth, slice):
        w = pyrtl.Input(bitwidth)
        with self.assertRaises(IndexError):
            _ = w[slice]

    def valid_slice(self, bitwidth, slice):
        w = pyrtl.Input(bitwidth)
        _ = w[slice]

    def test_wire_wo_bitwidth_fails(self):
        w = pyrtl.WireVector()
        with self.assertRaises(pyrtl.PyrtlError):
            _ = w[2]
        with self.assertRaises(pyrtl.PyrtlError):
            _ = w[3:5]

    def test_valid_indicies(self):
        self.valid_slice(4, 2)
        self.valid_slice(4, 0)
        self.valid_slice(4, -1)
        pyrtl.working_block().sanity_check()

    def test_valid_slices(self):
        self.valid_slice(8, slice(6))
        self.valid_slice(8, slice(1, 4))
        self.valid_slice(
            8, slice(1, 8)
        )  # Yes, supplying a end index out of bounds is valid python
        self.valid_slice(8, slice(1, 2, 2))
        self.valid_slice(8, slice(1, 4, 2))
        self.valid_slice(8, slice(7, 1, -2))
        self.valid_slice(8, slice(-2))
        self.valid_slice(8, slice(-6, -2, 3))
        pyrtl.working_block().sanity_check()

    def test_invalid_indicies(self):
        self.invalid_slice_index(4, 5)
        self.invalid_slice_index(4, -5)

    def test_invalid_slices(self):
        self.invalid_empty_slice(8, slice(1, 1))
        self.invalid_empty_slice(8, slice(7, 1))
        self.invalid_empty_slice(8, slice(-1, 1, 2))


class TestInput(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_no_connection_to_inputs(self):
        x = pyrtl.WireVector(1)
        y = pyrtl.Input(1)
        with self.assertRaises(pyrtl.PyrtlError):
            y <<= x


class TestRegister(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.r = pyrtl.Register(bitwidth=3)

    def test_connection_small(self):
        self.r.next <<= pyrtl.Const(1, bitwidth=3)

    def test_connection_unspec_width(self):
        self.r.next <<= pyrtl.Const(1)

    def test_connection_raw(self):
        self.r.next <<= 1

    def test_connection_large(self):
        self.r.next <<= pyrtl.Const(202)

    def test_register_assignment(self):
        with self.assertRaises(pyrtl.PyrtlError):
            self.r.next = 1

    def test_logic_operations(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = self.r or True

    def test_register_assignment_not_next(self):
        with self.assertRaises(pyrtl.PyrtlError):
            self.r <<= 1

    def test_connect_next(self):
        w = pyrtl.WireVector(bitwidth=1)
        with self.assertRaises(pyrtl.PyrtlError):
            w <<= self.r.next

    def test_next_logic_operations(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = self.r.next or True

    def test_reset_value_is_none(self):
        self.assertIsNone(self.r.reset_value)

    def test_reset_value_is_correct(self):
        r = pyrtl.Register(4, reset_value=1)
        self.assertEqual(r.reset_value, 1)

    def test_reset_value_as_string(self):
        r = pyrtl.Register(4, reset_value="2'd1")
        self.assertEqual(r.reset_value, 1)

    def test_invalid_reset_value_too_large(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.Register(4, reset_value=16)

    def test_invalid_reset_value_too_large_as_string(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.Register(4, reset_value="5'd16")

    def test_negative_reset_value(self):
        r = pyrtl.Register(4, reset_value=-4)
        self.assertEqual(pyrtl.val_to_signed_integer(r.reset_value, r.bitwidth), -4)

    def test_negative_reset_value_as_string(self):
        r = pyrtl.Register(4, reset_value="-4'd1")
        self.assertEqual(pyrtl.val_to_signed_integer(r.reset_value, r.bitwidth), -1)

    def test_invalid_negative_reset_value_as_string(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.Register(2, reset_value="-4'd1")

    def test_extending_negative_reset_value_as_string(self):
        r = pyrtl.Register(4, reset_value="-3'd3")
        self.assertEqual(pyrtl.val_to_signed_integer(r.reset_value, r.bitwidth), -3)

    def test_invalid_reset_value_not_an_integer(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.Register(4, reset_value="hello")


class TestConst(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_integers(self):
        self.check_const(1, 1, 1)
        self.check_const(5, 5, 3)
        self.check_const(1, 1, 5, bitwidth=5)
        self.check_const(0, 0b0, 1)
        self.check_const(0, 0b0, 1, signed=True)
        self.check_const(1, 0b01, 2, signed=True)
        self.check_const(2, 0b010, 3, signed=True)
        self.check_const(3, 0b011, 3, signed=True)
        self.check_const(4, 0b0100, 4, signed=True)
        self.check_const(5, 0b0101, 4, signed=True)

    def test_neg_integers(self):
        self.check_const(-1, 0b11111, 5, bitwidth=5)
        self.check_const(-2, 0b110, 3, bitwidth=3)
        self.check_const(-5, 0b1011, 4, bitwidth=4)
        self.check_const(-1, 0b1, 1, signed=True)
        self.check_const(-2, 0b10, 2, signed=True)
        self.check_const(-3, 0b101, 3, signed=True)
        self.check_const(-4, 0b100, 3, signed=True)
        self.check_const(-5, 0b1011, 4, signed=True)

    def test_too_big(self):
        self.assert_bad_const(5, 2)

    def test_invalid_bitwidth(self):
        self.assert_bad_const(1, 0)
        self.assert_bad_const(1, -1)

    def test_bad_neg_integers(self):
        # check that bitwidth is required
        self.assert_bad_const(-4)
        self.assert_bad_const(-4, 2)

    def test_string(self):
        self.check_const("1'1", 1, 1)
        self.check_const("5'3", 3, 5)
        self.check_const("5'b11", 3, 5)
        self.check_const("-5'b11", 29, 5)
        self.check_const("16'xff", 0xFF, 16)
        self.check_const("17'xff", 0xFF, 17)
        self.check_const("16'hff", 0xFF, 16)
        self.check_const("17'hff", 0xFF, 17)
        self.check_const("5'b011", 3, 5)
        self.check_const("5'b0_11", 3, 5)
        self.check_const("5'02_1", 21, 5)
        self.check_const("16'HFF", 0xFF, 16)

    def test_bad_string(self):
        self.assert_bad_const("1")
        self.assert_bad_const("-1")
        self.assert_bad_const("1bx")
        self.assert_bad_const("1ba")
        self.assert_bad_const("1'bx")
        self.assert_bad_const("1'z0")
        self.assert_bad_const("1'ba")
        self.assert_bad_const("1'b10")
        self.assert_bad_const("4'h12")
        self.assert_bad_const("-'h1")
        self.assert_bad_const("-2'b10")
        self.assert_bad_const("1'-b10")
        self.assert_bad_const("-1'b10")
        self.assert_bad_const("5'b111111'")
        self.assert_bad_const("'")
        self.assert_bad_const("'1")
        self.assert_bad_const("1'")

    def test_bool(self):
        self.check_const(True, 1, 1)
        self.check_const(False, 0, 1, bitwidth=1)

    def test_badbool(self):
        self.assert_bad_const(False, bitwidth=2)
        self.assert_bad_const(True, bitwidth=0)

    def test_badtype(self):
        self.assert_bad_const(pyrtl.Const(123))
        self.assert_bad_const([])

    def test_assignment(self):
        with self.assertRaises(pyrtl.PyrtlError):
            c = pyrtl.Const(4)
            c <<= 3

    def test_named(self):
        block = pyrtl.working_block()
        c = pyrtl.Const(20, name="archid")
        self.assertIn("archid", block.wirevector_by_name)
        self.assertIn(c, block.wirevector_set)
        self.assertEqual(c.val, 20)
        c.name = "vendorid"
        self.assertNotIn("archid", block.wirevector_by_name)
        self.assertIn("vendorid", block.wirevector_by_name)
        self.assertIn(c, block.wirevector_set)
        self.assertEqual(c.val, 20)

    def check_const(self, val_in, expected_val, expected_bitwidth, **kargs):
        c = pyrtl.Const(val_in, **kargs)
        self.assertEqual(c.val, expected_val)
        self.assertEqual(c.bitwidth, expected_bitwidth)

    def assert_bad_const(self, *args, **kwargs):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = pyrtl.Const(*args, **kwargs)


class TestOutput(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_assign_output(self):
        o = pyrtl.Output(1)
        w = pyrtl.WireVector(1)
        with self.assertRaises(pyrtl.PyrtlInternalError):
            w <<= o

    def test_log_op_output(self):
        o = pyrtl.Output(1)
        w = pyrtl.WireVector(1)
        with self.assertRaises(pyrtl.PyrtlInternalError):
            _ = w & o

    def test_slice_output(self):
        o = pyrtl.Output(2)
        with self.assertRaises(pyrtl.PyrtlInternalError):
            _ = o[0]


class TestKeepingCallStack(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    @classmethod
    def tearDownClass(cls):
        pyrtl.set_debug_mode(False)

    def test_no_call_stack(self):
        pyrtl.set_debug_mode(False)
        wire = pyrtl.WireVector()
        with self.assertRaises(AttributeError):
            _ = wire.init_call_stack

    def test_get_call_stack(self):
        pyrtl.set_debug_mode(True)
        wire = pyrtl.WireVector()
        call_stack = wire.init_call_stack
        self.assertIsInstance(call_stack, list)


class TestWrappedWireVector(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_attr(self):
        reg = pyrtl.Register(bitwidth=1, name="reg")
        wrapped_reg = pyrtl.wire.WrappedWireVector(reg)
        # Check __setattr__ forwarding.
        wrapped_reg.next <<= True

        # Check __getattr__ forwarding.
        self.assertEqual(wrapped_reg.bitwidth, 1)
        self.assertEqual(wrapped_reg.name, "reg")

        sim = pyrtl.Simulation()
        sim.step()
        self.assertEqual(sim.inspect("reg"), False)
        sim.step()
        self.assertEqual(sim.inspect("reg"), True)

    def test_ops(self):
        # Check special method forwarding.
        for a_val in range(16):
            for b_val in range(16):
                pyrtl.reset_working_block()

                a = pyrtl.Const(bitwidth=4, val=a_val)
                wrapped_a = pyrtl.wire.WrappedWireVector(a)
                b = pyrtl.Const(bitwidth=4, val=b_val)
                wrapped_b = pyrtl.wire.WrappedWireVector(b)

                self.assertEqual(hash(a), hash(wrapped_a))
                self.assertEqual(str(a), str(wrapped_a))
                self.assertEqual(repr(a), repr(wrapped_a))
                self.assertEqual(len(a), len(wrapped_a))

                a_and_b = pyrtl.WireVector(name="a_and_b", bitwidth=4)
                a_and_b <<= wrapped_a & wrapped_b
                a_rand_b = pyrtl.WireVector(name="a_rand_b", bitwidth=4)
                a_rand_b <<= wrapped_a.val & wrapped_b
                a_or_b = pyrtl.WireVector(name="a_or_b", bitwidth=4)
                a_or_b <<= wrapped_a | wrapped_b
                a_ror_b = pyrtl.WireVector(name="a_ror_b", bitwidth=4)
                a_ror_b <<= wrapped_a.val | wrapped_b
                a_xor_b = pyrtl.WireVector(name="a_xor_b", bitwidth=4)
                a_xor_b <<= wrapped_a ^ wrapped_b
                a_rxor_b = pyrtl.WireVector(name="a_rxor_b", bitwidth=4)
                a_rxor_b <<= wrapped_a.val ^ wrapped_b
                a_add_b = pyrtl.WireVector(name="a_add_b", bitwidth=5)
                a_add_b <<= wrapped_a + wrapped_b
                a_radd_b = pyrtl.WireVector(name="a_radd_b", bitwidth=5)
                a_radd_b <<= wrapped_a.val + wrapped_b
                a_sub_b = pyrtl.WireVector(name="a_sub_b", bitwidth=4)
                a_sub_b <<= wrapped_a - wrapped_b
                a_rsub_b = pyrtl.WireVector(name="a_rsub_b", bitwidth=4)
                a_rsub_b <<= wrapped_a.val - wrapped_b
                a_mul_b = pyrtl.WireVector(name="a_mul_b", bitwidth=8)
                a_mul_b <<= wrapped_a * wrapped_b
                a_rmul_b = pyrtl.WireVector(name="a_rmul_b", bitwidth=8)
                a_rmul_b <<= wrapped_a.val * wrapped_b
                a_lt_b = pyrtl.WireVector(name="a_lt_b", bitwidth=1)
                a_lt_b <<= wrapped_a < wrapped_b
                a_le_b = pyrtl.WireVector(name="a_le_b", bitwidth=1)
                a_le_b <<= wrapped_a <= wrapped_b
                a_eq_b = pyrtl.WireVector(name="a_eq_b", bitwidth=1)
                a_eq_b <<= wrapped_a == wrapped_b
                a_ne_b = pyrtl.WireVector(name="a_ne_b", bitwidth=1)
                a_ne_b <<= wrapped_a != wrapped_b
                a_gt_b = pyrtl.WireVector(name="a_gt_b", bitwidth=1)
                a_gt_b <<= wrapped_a > wrapped_b
                a_ge_b = pyrtl.WireVector(name="a_ge_b", bitwidth=1)
                a_ge_b <<= wrapped_a >= wrapped_b
                a_invert = pyrtl.WireVector(name="a_invert", bitwidth=4)
                a_invert <<= ~wrapped_a
                a_high = pyrtl.WireVector(name="a_high", bitwidth=2)
                a_high <<= wrapped_a[2:4]
                a_low = pyrtl.WireVector(name="a_low", bitwidth=2)
                a_low <<= wrapped_a[0:2]

                x = pyrtl.WireVector(name="x", bitwidth=4)
                wrapped_x = pyrtl.wire.WrappedWireVector(x)
                wrapped_x <<= a_val
                self.assertEqual(type(wrapped_x), pyrtl.wire.WrappedWireVector)

                sim = pyrtl.Simulation()
                sim.step()
                self.assertEqual(sim.inspect("a_and_b"), a.val & b.val)
                self.assertEqual(sim.inspect("a_rand_b"), a.val & b.val)
                self.assertEqual(sim.inspect("a_or_b"), a.val | b.val)
                self.assertEqual(sim.inspect("a_ror_b"), a.val | b.val)
                self.assertEqual(sim.inspect("a_xor_b"), a.val ^ b.val)
                self.assertEqual(sim.inspect("a_rxor_b"), a.val ^ b.val)
                self.assertEqual(sim.inspect("a_add_b"), a.val + b.val)
                self.assertEqual(sim.inspect("a_radd_b"), a.val + b.val)
                # Mask with 0xF to convert from signed to unsigned value.
                self.assertEqual(sim.inspect("a_sub_b"), (a.val - b.val) & 0xF)
                self.assertEqual(sim.inspect("a_rsub_b"), (a.val - b.val) & 0xF)
                self.assertEqual(sim.inspect("a_mul_b"), a.val * b.val)
                self.assertEqual(sim.inspect("a_rmul_b"), a.val * b.val)
                self.assertEqual(sim.inspect("a_lt_b"), a.val < b.val)
                self.assertEqual(sim.inspect("a_le_b"), a.val <= b.val)
                self.assertEqual(sim.inspect("a_eq_b"), a.val == b.val)
                self.assertEqual(sim.inspect("a_ne_b"), a.val != b.val)
                self.assertEqual(sim.inspect("a_gt_b"), a.val > b.val)
                self.assertEqual(sim.inspect("a_ge_b"), a.val >= b.val)
                # Mask with 0xF to convert from signed to unsigned value.
                self.assertEqual(sim.inspect("a_invert"), ~a.val & 0xF)
                self.assertEqual(sim.inspect("a_high"), a.val >> 2)
                self.assertEqual(sim.inspect("a_low"), a.val & 0x3)

                self.assertEqual(sim.inspect("x"), a.val)

    def test_conditional_assignment(self):
        # Check forwarding for __ior__, __enter__, __exit__.
        select = pyrtl.Input(name="select", bitwidth=1)
        wrapped_select = pyrtl.wire.WrappedWireVector(select)
        x = pyrtl.WireVector(name="x", bitwidth=4)
        wrapped_x = pyrtl.wire.WrappedWireVector(x)
        with pyrtl.conditional_assignment:
            with wrapped_select:
                wrapped_x |= 0xA
            with ~wrapped_select:
                wrapped_x |= 0xB
        self.assertEqual(type(wrapped_x), pyrtl.wire.WrappedWireVector)

        sim = pyrtl.Simulation()
        sim.step(provided_inputs={select: True})
        self.assertEqual(sim.inspect("x"), 0xA)
        sim.step(provided_inputs={select: False})
        self.assertEqual(sim.inspect("x"), 0xB)

    def test_exceptions(self):
        # Check forwarding for special methods that throw exceptions.
        a = pyrtl.Const(bitwidth=4, val=0xA)
        wrapped_a = pyrtl.wire.WrappedWireVector(a)
        b = pyrtl.Const(bitwidth=4, val=0xB)
        wrapped_b = pyrtl.wire.WrappedWireVector(b)

        with self.assertRaises(pyrtl.PyrtlError):
            bool(wrapped_a)
        with self.assertRaises(pyrtl.PyrtlError):
            wrapped_a &= wrapped_b
        with self.assertRaises(pyrtl.PyrtlError):
            wrapped_a ^= wrapped_b
        with self.assertRaises(pyrtl.PyrtlError):
            wrapped_a += wrapped_b
        with self.assertRaises(pyrtl.PyrtlError):
            wrapped_a -= wrapped_b
        with self.assertRaises(pyrtl.PyrtlError):
            wrapped_a *= wrapped_b
        with self.assertRaises(pyrtl.PyrtlError):
            wrapped_a << wrapped_b
        with self.assertRaises(pyrtl.PyrtlError):
            wrapped_a >> wrapped_b
        with self.assertRaises(pyrtl.PyrtlError):
            wrapped_a % wrapped_b


if __name__ == "__main__":
    unittest.main()
